# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import curses
from copy import deepcopy
# Find variable
from ..core.common import get_var
# Page class definition
from .jtopgui import Page
# Graphics elements
from .lib.colors import NColors
from .lib.smallbutton import HideButton
from .lib.common import plot_name_info, plot_dictionary
# Regex
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
COPYRIGHT_RE = re.compile(r""".*__copyright__ = ["'](.*?)['"]""", re.S)
EMAIL_RE = re.compile(r""".*__email__ = ["'](.*?)['"]""", re.S)


def plot_libraries(stdscr, pos_y, pos_x, libraries):
    libraries = deepcopy(libraries)
    opencv = libraries['OpenCV']
    opencv_cuda = libraries['OpenCV-Cuda']
    opencv_cuda_string = "YES" if opencv_cuda else "NO"
    color = NColors.green() if opencv_cuda else NColors.red()
    del libraries['OpenCV']
    del libraries['OpenCV-Cuda']
    # Plot Library
    libraries_size_y, libraries_size_x = plot_dictionary(stdscr, pos_y, pos_x, 'Libraries', libraries)
    # Plot OpenCV and CUDA
    opencv_string = opencv if opencv else "MISSING"
    opencv_color = curses.A_NORMAL if opencv else NColors.red()
    opencv_size_x = plot_name_info(stdscr, pos_y + libraries_size_y, pos_x + 1, 'OpenCV', opencv_string, color=opencv_color)
    len_opencv = opencv_size_x
    if opencv:
        try:
            stdscr.addstr(pos_y + libraries_size_y, pos_x + opencv_size_x + 1, " with CUDA:")
            stdscr.addstr(pos_y + libraries_size_y, pos_x + opencv_size_x + 13, opencv_cuda_string, color | curses.A_BOLD)
        except curses.error:
            pass
        len_opencv += len(opencv_cuda_string) + 15
    return libraries_size_y + 1, max(libraries_size_x, len_opencv)


def plot_hardware(stdscr, pos_y, pos_x, hardware, size_hardware_x):
    hardware = deepcopy(hardware)
    # Remove from list and add a button
    if 'Serial Number' in hardware:
        del hardware['Serial Number']
    # Plot all hardware data
    hardware_size_y, hardware_size_x = plot_dictionary(stdscr, pos_y, pos_x, 'Hardware', hardware, size=size_hardware_x)
    return hardware_size_y, hardware_size_x


class INFO(Page):

    def __init__(self, stdscr, jetson):
        super(INFO, self).__init__("INFO", stdscr, jetson)
        # Store serial number
        self._hide_serial_number = None
        hardware = self.jetson.board['hardware']
        if 'Serial Number' in hardware:
            if self.jetson.board['hardware']['Serial Number']:
                serial_number = hardware['Serial Number']
                self._hide_serial_number = HideButton(stdscr, 's', serial_number)

    def draw(self, key, mouse):
        """
            Write all environment variables
        """
        # Screen size
        _, width, first = self.size_page()
        start_pos = first + 3
        # Clear first line
        self.stdscr.move(0, 0)
        self.stdscr.clrtoeol()
        # Author info
        string_author = "jtop {version} - {copyright} [{email}]".format(version=get_var(VERSION_RE),
                                                                        copyright=get_var(COPYRIGHT_RE),
                                                                        email=get_var(EMAIL_RE))
        self.stdscr.addstr(first, 0, string_author, curses.A_BOLD)
        self.stdscr.addstr(first + 1, 0, "Website: https://rnext.it/jetson_stats", curses.A_BOLD)
        # Plot platform
        platform_size_y, platform_size_x = plot_dictionary(self.stdscr, start_pos, 1, 'Platform', self.jetson.board['platform'])
        # Plot libraries
        libraries_size_y, libraries_size_x = plot_libraries(self.stdscr, start_pos + platform_size_y + 1, 1, self.jetson.board['libraries'])
        # Plot hardware
        size_hardware_x = width - platform_size_x - 2
        # Plot serial number
        offset_y_sn = 0
        if 'Serial Number' in self.jetson.board['hardware']:
            if self.jetson.board['hardware']['Serial Number']:
                try:
                    self.stdscr.addstr(start_pos, 1 + platform_size_x + 1, "Serial Number:", curses.A_BOLD)
                except curses.error:
                    pass
                self._hide_serial_number.update(start_pos, 1 + platform_size_x + 16, key=key, mouse=mouse)
                offset_y_sn = 1
        hardware_size_y, hardware_size_x = plot_hardware(self.stdscr,
                                                         start_pos + offset_y_sn,
                                                         1 + platform_size_x + 1,
                                                         self.jetson.board['hardware'], size_hardware_x)
        hardware_size_y += offset_y_sn
        # Plot interfaces
        interfaces = self.jetson.local_interfaces["interfaces"]
        hostname = self.jetson.local_interfaces["hostname"]
        max_size_x = max(platform_size_x, libraries_size_x)
        plot_name_info(self.stdscr, start_pos + hardware_size_y + 1, 2 + max_size_x, "Hostname", hostname)
        interfaces_size_y, interfaces_size_x = plot_dictionary(self.stdscr, start_pos + hardware_size_y + 2, 2 + max_size_x, 'Interfaces', interfaces)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import time
from threading import Thread, Event
# Logging
import logging
# Create logger
logger = logging.getLogger(__name__)

TIMER_READER_MIN_SLEEP = 0.05


class TimerReader:

    def __init__(self, callback):
        self._stop_event = Event()
        # Initialize callback
        self._callback = callback
        # Error message from thread
        self._error = None
        # Define Thread
        self._thread = None

    def _timer_callback(self, interval, stop_event):
        logger.debug("jtop timer start at {interval}s".format(interval=interval))
        try:
            while stop_event.is_set():
                start = time.time()
                # Callback function
                self._callback()
                # Measure timer_callback sleep time
                delta = time.time() - start
                # Start to sleep
                if interval > delta:
                    time.sleep(interval - delta)
        except (KeyboardInterrupt, SystemExit):
            logger.info("KeyboardInterrupt or SystemExit, exit timer_reader thread")
        except Exception as e:
            logger.fatal("Exception in 'timer_reader thread': {}".format(e))
            # Store error message
            self._error = sys.exc_info()
        logger.debug("jtop timer stopped")

    def open(self, interval=0.5):
        # Catch exception if exist
        self._error_status()
        # Check if not running
        if self._thread is not None:
            return False
        # Check if thread or process exist
        self._stop_event.set()
        # Start thread Service client
        self._thread = Thread(target=self._timer_callback, args=(interval, self._stop_event, ))
        self._thread.start()
        return True

    def close(self, timeout=None):
        # Catch exception if exist
        self._error_status()
        # Check if thread and process are already empty
        self._stop_event.clear()
        if self._thread is not None:
            self._thread.join(timeout)
            self._thread = None
        return True

    def _error_status(self):
        # Catch exception if exist
        if not self._error:
            return
        # Extract exception and raise
        ex_type, ex_value, tb_str = self._error
        ex_value.__traceback__ = tb_str
        raise ex_value
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import logging
import threading
# Launch command
import subprocess as sp
# Load queue library for python 2 and python 3
try:
    import queue
except ImportError:
    import Queue as queue  # pyright: ignore[reportMissingImports]
# Create logger
logger = logging.getLogger(__name__)
EXTRA_TIMEOUT = 1.0
# Reference:
# https://eli.thegreenplace.net/2017/interacting-with-a-long-running-child-process-in-python/
# https://stackoverflow.com/questions/37942022/returncode-of-popen-object-is-none-after-the-process-is-terminated/42376107
# https://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python
# https://docs.python.org/3/tutorial/errors.html
# https://stackoverflow.com/questions/10756383/timeout-on-subprocess-readline-in-python
# https://stackoverflow.com/questions/3733270/python-subprocess-timeout


class Command(object):

    class CommandException(Exception):

        def __init__(self, message, errno):
            self.message = message
            self.errno = errno

        def __str__(self):
            return "[errno:{errno}] {message}".format(message=self.message, errno=self.errno)

    class TimeoutException(CommandException):

        def __init__(self):
            super(Command.TimeoutException, self).__init__("Process does not replied in time", -1)

    @staticmethod
    def run_command(command, repeat=5, timeout=2, input=None):
        cmd = Command(command, input=input)
        for idx in range(repeat):
            try:
                return cmd(timeout=timeout)
            except Command.TimeoutException as error:
                logger.error("[{idx}] {error}".format(idx=idx, error=error))
        raise Command.CommandException("Error to start {command}".format(command=command), -2)

    def __init__(self, command, input=None):
        self.process = None
        self.command = command
        self.input = input

    def __call__(self, timeout=None):
        def target(out_queue, err_queue):
            # Run process
            try:
                # https://stackoverflow.com/questions/33277452/prevent-unexpected-stdin-reads-and-lock-in-subprocess
                if self.input is not None:
                    self.process = sp.Popen(self.command, stdout=sp.PIPE, stderr=sp.PIPE, stdin=sp.PIPE, preexec_fn=os.setsid)
                    # Write input
                    stdout, _ = self.process.communicate(bytes(self.input, 'utf-8'))
                    self.process.stdin.close()
                    self.process.wait()
                    # Read lines output
                    for line in stdout.splitlines():
                        try:
                            line = line.decode('utf-8')
                            line = str(line.strip())
                        except UnicodeEncodeError:
                            line = line.encode('ascii', 'ignore').decode('ascii')
                        out_queue.put(line)
                else:
                    self.process = sp.Popen(self.command, stdout=sp.PIPE, stderr=sp.PIPE, stdin=open(os.devnull), preexec_fn=os.setsid)
                    # Read lines output
                    for line in iter(self.process.stdout.readline, b''):
                        try:
                            line = line.decode('utf-8')
                            line = str(line.strip())
                        except UnicodeEncodeError:
                            line = line.encode('ascii', 'ignore').decode('ascii')
                        out_queue.put(line)
                    # Close and terminate
                    self.process.stdout.close()
                    self.process.wait()
            except Exception:
                # Store error message
                err_queue.put(sys.exc_info())
        # Initialize lists
        is_timeout = False
        out_queue = queue.Queue()
        err_queue = queue.Queue()
        thread = threading.Thread(target=target, args=(out_queue, err_queue, ))
        thread.start()
        # Wait timeout process
        thread.join(timeout)
        if thread.is_alive():
            logger.error('Terminating process: {command}'.format(command=self.command))
            if self.process is not None:
                self.process.terminate()
            thread.join(timeout=EXTRA_TIMEOUT)
            logger.warning('Process terminated: {command}'.format(command=self.command))
            is_timeout = True
        # Read the output
        # Extract exception and raise
        if not err_queue.empty():
            ex_type, ex_value, tb_str = err_queue.get()
            ex_value.__traceback__ = tb_str
            raise ex_value
        if is_timeout:
            raise Command.TimeoutException()
        if self.process.returncode != 0:
            raise Command.CommandException('Error process:', self.process.returncode)
        return list(out_queue.queue)

    def communicate(self, timeout=None):
        self.__call__(timeout=timeout)
# EOF



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Shell information
# os.environ['SHELL']

from copy import deepcopy
# jtop variable
from .core.gpu import get_raw_igpu_devices
from .core.hardware import get_platform_variables
from .core.jetson_variables import get_jetson_raw_output
from .terminal_colors import bcolors


def engine_gui(repository, hardware, version):
    hardware = deepcopy(hardware)
    del hardware['Serial Number']
    module = hardware['Module']
    # Title
    title = "GUI compact engine page missing for [{module}]".format(module=module)
    # Template
    template = "engine-gui.md"
    # body
    body = "Please add a new jtop engine compact page for\n\n"
    body += "### Board\n\n"
    for name, value in hardware.items():
        if not value:
            value = "**MISSING**"
        body += " - {name}: {value}\n".format(name=name, value=value)
    body += "\n### jetson-stats\n\n"
    body += " - Version: " + version + "\n"
    # Print all raw output
    body += "\n<!-- Please attach a screenshot page from jtop Engine page -->\n"
    body += "### Screenshot engine page\n\n"
    body += "Screenshot page from jtop engine page attached"
    # Make url
    url = make_issue(repository, title, body=body, labels="GUI,missing", template=template)
    # message shell
    message = "Module \"{module}\" missing in jtop GUI".format(module=module)
    hyperlink(message, url, "open a Github issue & {attach} a engine page screenshot".format(attach=bcolors.bold('attach')))


def jetpack_missing(repository, hardware, version):
    l4t = hardware['L4T']
    # Title
    title = "jetson-stats not supported for [L4T {l4t}]".format(l4t=l4t)
    # Template
    template = "jetpack-missing.md"
    # Body
    body = "Please update jetson-stats with new jetpack\n\n"
    body += "### Linux for Tegra\n\n"
    body += " - L4T: " + l4t + "\n\n"
    body += "### Jetson-Stats\n\n"
    body += " - Version: " + version + "\n"
    # Make url
    url = make_issue(repository, title, body=body, labels="Jetpack,missing", template=template)
    # message shell
    message = "jetson-stats not supported for [L4T {l4t}]".format(l4t=hardware['L4T'])
    hyperlink(message, url, "open a Github issue")


def get_hardware_log():
    # Print platform
    platform = get_platform_variables()
    body = "--------------------- PLATFORM -------------------------\n"
    for name, value in platform.items():
        body += "{name}: {value}\n".format(name=name, value=value)
    # Print all jetson raw output
    raw_output = get_jetson_raw_output()
    body += "-------------------- JETSON RAW OUTPUT -----------------\n"
    for name, value in raw_output.items():
        body += "------------------\n"
        body += "Path: {name}\n{value}\n".format(name=name, value=value)
    # Print device list
    raw_output = get_raw_igpu_devices()
    body += "\n-------------------- IGPU OUTPUT ---------------------\n"
    for path, value in raw_output.items():
        body += "------------------\n"
        body += "Path: {path}\n{value}\n".format(path=path, value=value)
    return body


def hardware_missing(repository, hardware, version):
    hardware = deepcopy(hardware)
    del hardware['Serial Number']
    # Title
    if 'P-Number' in hardware:
        title = "Hardware Missing: {pnumber}".format(pnumber=hardware.get('P-Number', ''))
    else:
        title = "Hardware Missing".format()
    # Template
    template = "hardware-missing.md"
    # Body
    body = "Please update jetson-stats with this board\n\n"
    body += "### Board\n\n"
    for name, value in hardware.items():
        if not value:
            value = "**MISSING**"
        body += " - {name}: {value}\n".format(name=name, value=value)
    body += "\n### Jetson-Stats\n\n"
    body += " - Version: " + version + "\n"
    # Print all raw output
    body += "\n<!-- Please attach the output from: jtop --error-log -->\n"
    body += "### RAW Data\n\n"
    body += "File from `--error-log` attached"
    # Make url
    url = make_issue(repository, title, body=body, labels="Hardware,missing", template=template)
    # message shell
    message = "jtop not support this hardware"
    hyperlink(message, url, "open a Github issue & {attach} file from: jtop --error-log".format(attach=bcolors.bold('attach')))


def hyperlink(message, url, text):
    # Reference:
    # 1. http://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html
    # 2. https://stackoverflow.com/questions/40419276/python-how-to-print-text-to-console-as-hyperlink
    # 3. https://purpleidea.com/blog/2018/06/29/hyperlinks-in-gnome-terminal/
    # 4. https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda
    # 5. https://stackoverflow.com/questions/44078888/clickable-html-links-in-python-3-6-shell
    # Print starting message
    print("[{status}] {message}".format(status=bcolors.warning(), message=message))
    print("  Please, try: {bold}sudo pip3 install -U jetson-stats{reset} or".format(bold=bcolors.BOLD, reset=bcolors.ENDC))
    # Generate hyperlink for shell
    # Check type of shell
    try:
        link = u'\u001b]8;;{url}\u001b\\{text}\u001b]8;;\u001b\\ (press CTRL + Click)'.format(url=url, text=text)
    except UnicodeDecodeError:
        link = "{text}".format(text=text)
    print("  {link}".format(link=link))


def make_issue(repository, title, body="", labels="", template=""):
    # Reference
    # https://help.github.com/en/enterprise/2.17/user/github/managing-your-work-on-github/about-automation-for-issues-and-pull-requests-with-query-parameters
    # https://krypted.com/utilities/html-encoding-reference/
    title = title.replace(" ", "+").replace("[", "%5B").replace("]", "%5D")
    url = "{repository}/new?title={title}".format(repository=repository, title=title)
    if labels:
        url += "&labels=" + labels
    if template:
        url += "&template=" + template
    if body:
        body = body.replace(" ", "+").replace("#", "%23").replace("\n", "%0A").replace("*", "%2A")
        url += "&body=" + body
    return url
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import time
from threading import Thread
# Logging
import logging
# Launch command
from .command import Command
# Import exceptions
from .exceptions import JtopException
# Fix connection refused for python 2.7
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

# Create logger
logger = logging.getLogger(__name__)
# Regular expressions
POWER_MODEL_DEFAULT_RE = re.compile(r'PM_CONFIG: DEFAULT=(?P<name>\w+)\((?P<id>\d+)\)')
TPC_POWER_GATING_RE = re.compile(r'TPC_POWER_GATING TPC_PG_MASK (?P<mask>\w+)')
POWER_MODEL_RE = re.compile(r'POWER_MODEL: ID=(?P<id>\d+) NAME=(?P<name>\w+)')
NV_POWER_MODE_RE = re.compile(r'NV Power Mode: (?P<name>\w+)')

COMMAND_TIMEOUT = 4.0
NVP_TIMER_WAIT_JETSON_CLOCKS = 0.5
NVP_COUNTER_ALIVE_JETSON_CLOCKS = 5


def nvpmodel_decode():
    default = {}
    nvpm = {}
    nvpm_masks = {}
    nvpmodel_p = Command(['nvpmodel', '-p', '--verbose'])
    lines = nvpmodel_p(timeout=COMMAND_TIMEOUT)
    # Decode lines
    for line in lines:
        # Search default value
        match = re.search(POWER_MODEL_DEFAULT_RE, line)
        if match:
            parsed_line = match.groupdict()
            default = {'name': parsed_line['name'], 'id': int(parsed_line['id'])}
        # Search configuration NVPmodel
        match = re.search(POWER_MODEL_RE, line)
        # if match extract name and number
        if match:
            parsed_line = match.groupdict()
            # Extract id and name
            mode_id = int(parsed_line['id'])
            # Save in nvpm list
            nvpm[mode_id] = parsed_line['name']
        # Search TPC Power Gating
        match = re.search(TPC_POWER_GATING_RE, line)
        if match:
            parsed_line = match.groupdict()
            # Extract save mask in nvpm_masks list with the same id
            nvpm_masks[mode_id] = parsed_line['mask']
    # Make a list
    nvpm_masks_list = list(nvpm_masks.values()) if nvpm_masks else [True] * len(nvpm)
    return default, list(nvpm.values()), nvpm_masks_list


def nvpmodel_query():
    """ Read nvpmodel to know the status of the board """
    try:
        nvpmodel_p = Command(['nvpmodel', '-q'])
        lines = nvpmodel_p(timeout=COMMAND_TIMEOUT)
    except FileNotFoundError:
        raise JtopException("nvpmodel doesn't exist")
    except Command.CommandException:
        raise JtopException("nvpmodel command unavailable")
    # Extract lines
    for idx, line in enumerate(lines):
        # Search configuration NVPmodel
        match = NV_POWER_MODE_RE.search(line)
        # if match extract name and number
        if match:
            parsed_line = match.groupdict()
            # Extract NV Power Mode
            return {'name': parsed_line['name'], 'id': int(lines[idx + 1])}


def set_nvpmodel_level(level, force):
    """ Set nvpmodel to a new status """
    # Set the new nvpmodel status
    try:
        # If there are no errors return the NV Power mode
        if force:
            lines = Command.run_command(['nvpmodel', '-m', str(level)], repeat=5, timeout=COMMAND_TIMEOUT, input='YES')
        else:
            lines = Command.run_command(['nvpmodel', '-m', str(level)], repeat=5, timeout=COMMAND_TIMEOUT)
        # Check if error is in vector
        for line in lines:
            if "NVPM ERROR" in line:
                text = "\n".join(lines)
                logger.error("Error to set nvpmodel {level}. Message:\n{text}".format(level=level, text=text))
                return False
    except Command.TimeoutException:
        logger.error("nvpmodel does not reply in time")
        return False
    except Command.CommandException:
        logger.error("nvpmodel command exception")
        return False
    return True


class NVPModel(object):
    """
    The NVP Model control voltage regulators, and power tree to optimize power efficiency.
    It supports three optimized power budgets, such as 10 watts, 15 watts, and 30 watts.
    For each power budget, several configurations are possible with various CPU frequencies and number of cores online.

    Capping the memory, CPU, and GPU frequencies, and number of online CPU, GPU TPC, DLA and PVA cores at a prequalified level confines the module to the target mode.

    This method simplify in a set of functions and variables this controller.

    .. code-block:: python

        with jtop() as jetson:
            if jetson.ok():
                # Read current nvpmodel name
                print(jetson.nvpmodel)
                # List of all nvpmodel available
                models = jetson.nvpmodel.models
                print(models)
                # You can write a string for a name
                jetson.nvpmodel = models[0]
                # or an the ID name is also allowed
                jetson.nvpmodel = 0

    You can also increase/decrease the ID

    .. code-block:: python

        with jtop() as jetson:
            if jetson.ok():
                jetson.nvpmodel += 1
                # or
                jetson.nvpmodel = jetson.nvpmodel + 1

    Read how to use in :py:attr:`~jtop.jtop.nvpmodel`

    .. admonition:: Reference

        #. `NVP Model - Jetson TX/Nano <https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-283/Tegra%20Linux%20Driver%20Package%20Development%20Guide/power_management_tx2.html#wwpID0E0AM0HA>`_
        #. `NVP Model - Jetson Xavier <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonXavierNxSeriesAndJetsonAgxXavierSeries.html#supported-modes-and-power-efficiency>`_
        #. `NVP Model - Jetson Orin <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#supported-modes-and-power-efficiency>`_
    """  # noqa

    def __init__(self, controller, nvpmodel):
        self._controller = controller
        self._nvp_models = nvpmodel['models']
        self._nvp_default = nvpmodel['default']
        self._status = []
        self._running = False
        self._nvpmodel_now = {}

    def _update(self, nvp_status):
        self._status = nvp_status['status']
        self._running = nvp_status['thread']
        self._nvpmodel_now = nvp_status['model']

    def is_running(self):
        """
        Return the status of the NVP Model service, if **True** the service is already running.

        :return: status nvpmodel service
        :rtype: bool
        """
        return self._running

    @property
    def id(self):
        """
        Read che current nvpmodel ID status selected

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read current nvpmodel ID
                    print(jetson.nvpmodel.id)

        :return: nvpmodel ID
        :rtype: int
        """
        return self._nvpmodel_now['id']

    @property
    def name(self):
        """
        Read che current nvpmodel ID **name** selected

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read current nvpmodel name
                    print(jetson.nvpmodel.name)
                    # same of
                    print(jetson.nvpmodel)

        :return: nvpmodel ID **name**
        :rtype: str
        """
        return self._nvpmodel_now['name']

    @property
    def status(self):
        """
        Return a list with a NVP model status. If when you set an NVP Model, something going wrong, you will read here the output

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read current nvpmodel status
                    print(jetson.nvpmodel.status)

        Output

        .. code-block:: python
            :class: no-copybutton

            # [True True True True ... True]


        :return: Status NVP Models
        :rtype: list
        """
        return self._status

    def set_nvpmodel_id(self, nvpmodel_id, force=False):
        """
        Set a new NVP model by ID. The list of nvpmodel available is with :py:func:`~get_all_nvpmodels` or :py:attr:`~models`.

        .. danger::

            If you set **force=True** the board will be rebooted immediately without any warning.

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Set a new NVP Model by ID
                    jetson.nvpmodel.set_nvpmodel_id(0, force=False)
                    # same of
                    jetson.nvpmodel = 0

        :param nvpmodel_id: nvpmodel ID
        :type nvpmodel_id: int
        :param force: Force set nvpmodel (can reboot your board), defaults to False
        :type force: bool, optional
        :raises ValueError: the variable is not an int
        :raises JtopException: nvpmodel name doesn't' exist
        """
        if not isinstance(nvpmodel_id, int):
            raise ValueError("Use an int")
        # Check if ID is in list
        if nvpmodel_id < 0 or nvpmodel_id > len(self._nvp_models) - 1:
            raise JtopException("NV Power Model ID {id} does not exists! Range [0, {max}]".format(id=nvpmodel_id, max=len(self._nvp_models) - 1))
        if nvpmodel_id == self._nvpmodel_now['id'] and not force:
            return
        # Send new nvpmodel
        self._controller.put({'nvp': {'id': nvpmodel_id, 'force': force}})

    def get_nvpmodel_id(self):
        """
        Read che current nvpmodel ID status selected

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read current nvpmodel ID
                    print(jetson.nvpmodel.get_nvpmodel_id())
                    # same of
                    print(jetson.nvpmodel.id)

        :return: nvpmodel ID
        :rtype: int
        """
        return self._nvpmodel_now['id']

    def get_all_nvpmodels(self):
        """
        All nvp models available for this board

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # list of NVP models available
                    print(jetson.nvpmodel.get_all_nvpmodels())
                    # same of
                    print(jetson.nvpmodel.models)

        :return: List os string with all nvpmodels name
        :rtype: list
        """
        return self._nvp_models

    @property
    def models(self):
        """
        All nvp models available for this board

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # list of NVP models available
                    print(jetson.nvpmodel.models)

        :return: List os string with all nvpmodels name
        :rtype: list
        """
        return self._nvp_models

    def set_nvpmodel_name(self, nvpmodel_name, force=False):
        """
        Set a new NVP model by name. The list of nvpmodel available is with :py:func:`~get_all_nvpmodels` or :py:attr:`~models`.

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Set a new NVP Model by name
                    jetson.nvpmodel.set_nvpmodel_name("15W", force=False)
                    # same of
                    jetson.nvpmodel = "15W"

        :param nvpmodel_name: name nvpmodel
        :type nvpmodel_name: str
        :param force: Force set nvpmodel (can reboot your board), defaults to False
        :type force: bool, optional
        :raises ValueError: the variable is not a string
        :raises JtopException: nvpmodel name doesn't' exist
        """
        if not isinstance(nvpmodel_name, str):
            raise ValueError("Use a string")
        if nvpmodel_name not in self._nvp_models:
            raise JtopException("NV Power Model {name} does not exists! Check all NVPmode available".format(name=nvpmodel_name))
        if nvpmodel_name == self._nvpmodel_now['name'] and not force:
            return
        # Convert in nvpmodel id and send
        nvpmodel_id = self._nvp_models.index(nvpmodel_name)
        # Send new nvpmodel
        self._controller.put({'nvp': {'id': nvpmodel_id, 'force': force}})

    def get_nvpmodel_name(self):
        """
        Read che current nvpmodel ID **name** selected

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read current nvpmodel ID
                    print(jetson.nvpmodel.get_nvpmodel_name())
                    # same of
                    print(jetson.nvpmodel.name)
                    # or
                    print(jetson.nvpmodel)

        :return: nvpmodel ID **name**
        :rtype: str
        """
        return self._nvpmodel_now['name']

    def get_default(self):
        """
        Return a :py:class:`dict` with the default status for this board

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    default = jetson.nvpmodel.get_default()
                    print(default)

        output

        .. code-block:: python
            :class: no-copybutton

            # {'name': '15W', 'id': 0}

        :return: default NVP model
        :rtype: dict
        """
        return self._nvp_default

    def __add__(self, number):
        return self._nvpmodel_now['id'] + number

    def __radd__(self, number):
        return self._nvpmodel_now['id'] + number

    def __sub__(self, number):
        return self._nvpmodel_now['id'] - number

    def __rsub__(self, number):
        return self._nvpmodel_now['id'] - number

    def __iadd__(self, number):
        return self._nvpmodel_now['id'] + number

    def __isub__(self, number):
        return self._nvpmodel_now['id'] - number

    def __eq__(self, other):
        if isinstance(other, int):
            return self._nvpmodel_now['id'] == other
        elif isinstance(other, str):
            return self._nvpmodel_now['name'] == other
        elif isinstance(other, NVPModel):
            return self._nvpmodel_now['id'] == other._nvpmodel_now['id']
        else:
            return False

    def __repr__(self):
        return self._nvpmodel_now['name']


class NVPModelService(object):

    def __init__(self, jetson_clocks):
        self._is_nvpmodel = True
        # Initialize thread
        self._nvp_mode_set_thread = None
        # Initialize jetson_clocks config
        self._jetson_clocks = jetson_clocks
        try:
            # Read all NVP modes and masks available for this board
            self._default, self._nvp_models, self._nvp_masks = nvpmodel_decode()
            # Read current nvpmodel
            self._nvpmodel_now = nvpmodel_query()
            logger.info("nvpmodel running in [{id}]{name} - Default: {default}".format(
                name=self._nvpmodel_now['name'],
                id=self._nvpmodel_now['id'],
                default=self._default['id']))
            # Decode current mask
            current_mask = self._nvp_masks[self._nvpmodel_now['id']]
            # list of all nvpmodel status that can be changed from the current
            self._nvp_status = [current_mask == mask for idx, mask in enumerate(self._nvp_masks)]
        except (OSError, Command.CommandException):
            self._is_nvpmodel = False
            logger.warning("nvpmodel not available")

    def exists(self):
        return self._is_nvpmodel

    def get_all_nvpmodels(self):
        return self._nvp_models

    def _thread_set_nvp_model(self, nvpmodel_id, force):
        if not self._jetson_clocks.exists():
            # Set NV Power Mode
            status = set_nvpmodel_level(nvpmodel_id, force)
            # Update status
            self._nvp_status[nvpmodel_id] = status
            return
        if not self._jetson_clocks.is_config():
            # Set NV Power Mode
            status = set_nvpmodel_level(nvpmodel_id, force)
            # Update status
            self._nvp_status[nvpmodel_id] = status
            return
        # Otherwise disable the jetson_clocks
        old_status = self._jetson_clocks.get_enable()
        # Switch off jetson_clocks if is running
        if old_status:
            self._jetson_clocks.set_enable(False)
            # Check jetson_clocks is off
            counter = 0
            while self._jetson_clocks.get_enable():
                logger.info("[{counter}] NVPmodel waiting jetson_clocks setting off. Sleep for {time}s".format(
                    counter=counter, time=NVP_TIMER_WAIT_JETSON_CLOCKS))
                time.sleep(NVP_TIMER_WAIT_JETSON_CLOCKS)
                if counter > NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                    logger.error("NVPmodel jetson_clocks didn't change status in time")
                    break
                counter += 1
            if counter <= NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                logger.info("NVPmodel has switched off jetson_clocks")
        # Set NV Power Mode
        status = set_nvpmodel_level(nvpmodel_id, force)
        # Update status
        self._nvp_status[nvpmodel_id] = status
        # Status message
        if status:
            logger.info("NVPmodel started ID {id}".format(id=nvpmodel_id))
        else:
            logger.error("Error to set NVPmodel ID {id}".format(id=nvpmodel_id))
        # Switch off jetson_clocks if is running
        if old_status:
            self._jetson_clocks.set_enable(True)
            # Check jetson_clocks is off
            counter = 0
            while not self._jetson_clocks.get_enable():
                logger.info("[{counter}] NVPmodel waiting jetson_clocks setting on. Sleep for {time}s".format(
                    counter=counter, time=NVP_TIMER_WAIT_JETSON_CLOCKS))
                time.sleep(NVP_TIMER_WAIT_JETSON_CLOCKS)
                if counter > NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                    logger.error("NVPmodel jetson_clocks didn't change status in time")
                    break
                counter += 1
            if counter <= NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                logger.info("NVPmodel has switched on jetson_clocks")

    def set_nvpmodel_id(self, nvpmodel_id, force):
        if self.is_running():
            return False
        # Get current NV Power Mode
        old_nvp_mask = self._nvp_masks[self.get_nvpmodel_id()]
        nvp_mask = self._nvp_masks[nvpmodel_id]
        if force:
            logger.warning("Force set nvpmodel {nvpmodel_id}".format(nvpmodel_id=nvpmodel_id))
        else:
            if nvp_mask != old_nvp_mask:
                logger.error(
                    "The new nvpmodel {nvpmodel_id} has a different mask {nvp_mask}, is not compatible".format(
                        nvpmodel_id=nvpmodel_id, nvp_mask=nvp_mask))
                return False
        # Start thread Service client
        self._nvp_mode_set_thread = Thread(target=self._thread_set_nvp_model, args=(nvpmodel_id, force))
        # self._thread.daemon = True
        self._nvp_mode_set_thread.start()

    def get_nvpmodel_id(self):
        # Read current nvpmodel
        status = self.get_status()
        return status['model']['id']

    def get_nvpmodel_name(self):
        # Read current nvpmodel
        status = self.get_status()
        return status['model']['name']

    def set_default(self):
        self.set_nvpmodel_id(self._default['id'], False)

    def get_default(self):
        return self._default

    def is_running(self):
        if self._nvp_mode_set_thread is None:
            return False
        # https://github.com/fkie/multimaster_fkie/issues/149
        # https://docs.python.org/2.7/library/threading.html#threading.Thread.is_alive
        return self._nvp_mode_set_thread.is_alive()

    def close(self):
        # If jetson_clocks doesn't exist skip
        if not self.exists():
            return
        # If there are no thread running skip
        if self._nvp_mode_set_thread is None:
            return
        if self._nvp_mode_set_thread.is_alive():
            logger.warning("Wait switch off nvpmodel thread")
            self._nvp_mode_set_thread.join(COMMAND_TIMEOUT)

        self._nvp_mode_set_thread.join()

    def get_status(self):
        running = self.is_running()
        # If thread is not running update status
        if not running:
            self._nvpmodel_now = nvpmodel_query()
        return {
            'status': self._nvp_status,
            'thread': running,
            'model': self._nvpmodel_now,
        }
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
from .smallbutton import ButtonList
# Gui refresh rate
GUI_REFRESH = 1000 // 20


class DialogWindow(object):
    def __init__(self, title, text, on_click, buttons, width=44, height=6):
        self.dialog_width = width
        self.dialog_height = height
        self.title = title
        self.text = text
        self.on_click = on_click
        self.buttons = buttons
        self.enable_dialog_window = False
        self._dialog_win = None
        self.info = {}

    def enable(self, title="", info={}):
        if title:
            self.title = title
        self.info = info
        self.enable_dialog_window = True

    def disable(self):
        self.enable_dialog_window = False

    def show(self, stdscr, key, mouse):
        if self.enable_dialog_window:
            self._draw(stdscr, key, mouse)

    def _draw(self, stdscr, key, mouse):
        # Refresh the window to show the changes
        height, width = stdscr.getmaxyx()
        dialog_y, dialog_x = (height - self.dialog_height) // 2, (width - self.dialog_width) // 2
        self._dialog_win = curses.newwin(self.dialog_height, self.dialog_width, dialog_y, dialog_x)
        # Create a list of buttons
        self._buttons_profile = ButtonList(self._dialog_win, self._on_click, self.buttons, info=self.info, linear=True)
        # Add a border around the window
        self._dialog_win.border()
        # Add the title and the text
        self._dialog_win.addstr(1, 2, self.title, curses.A_BOLD)
        self._dialog_win.addstr(2, 2, self.text)
        # Add the buttons
        align_mouse = (mouse[0] - dialog_x, mouse[1] - dialog_y) if mouse else ()
        self._buttons_profile.update(4, 2, key, align_mouse, "")
        # Refresh the window to show the changes
        self._dialog_win.refresh()
        self._dialog_win.timeout(GUI_REFRESH)

    def _on_click(self, info, selected):
        self.on_click(info, selected)
        self.enable_dialog_window = False
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
# Page class definition
from .jtopgui import Page
from .lib.chart import Chart
from .lib.colors import NColors
from .lib.common import unit_to_string
from .lib.linear_gauge import freq_gauge, basic_gauge


def cpu_gauge(stdscr, idx, cpu, pos_y, pos_x, _, size_w):
    # online status
    online = cpu['online'] if 'online' in cpu else True
    # name cpu - workararound for TOTAL cpu
    name = cpu['name'] if 'name' in cpu else str(idx + 1) + (" " if idx < 9 else "")
    # Plot values
    values = [
        (cpu['user'], NColors.green()),
        (cpu['nice'], NColors.yellow()),
        (cpu['system'], NColors.red()),
    ] if online else []
    # Draw gauge
    data = {
        'name': name,
        'color': NColors.cyan(),
        'online': online,
        'values': values,
        # 'mleft': cpu.get('governor', '').capitalize(),
    }
    if size_w < 16:
        basic_gauge(stdscr, pos_y, pos_x, size_w - 1, data)
        return
    elif 'freq' in cpu:
        # Draw current frequency
        curr_string = unit_to_string(cpu['freq']['cur'], 'k', 'Hz')
        stdscr.addstr(pos_y, pos_x + size_w - 6, curr_string, NColors.italic())
    # Draw gauge
    basic_gauge(stdscr, pos_y, pos_x, size_w - 8, data)


def cpu_grid(stdscr, list_cpu, print_cpu, start_y, start_x, size_height=0, size_width=0):
    num_cpu = len(list_cpu)
    if num_cpu == 0:
        return size_height, size_width, 0, 0
    size_columns = 4 if num_cpu > 6 else 2
    # Measure size rows and columns
    size_rows = int(num_cpu / size_columns) + bool((num_cpu / size_columns) % 1)
    size_columns = int(num_cpu / size_rows) + bool((num_cpu / size_rows) % 1)
    # Measure step height and width
    step_height = int(round(size_height / size_rows)) if size_height > 0 else 1
    step_width = int(size_width / size_columns) if size_width > 0 else 1
    # Build Grid
    idx_row = 0
    idx_column = 0
    for idx, cpu in enumerate(list_cpu):
        # Check row index
        if idx_row >= size_rows:
            idx_row = 0
            idx_column += 1
        # Get CPU in grid
        try:
            print_cpu(stdscr, idx, cpu, start_y + idx_row * step_height, start_x + idx_column * step_width, step_height - 1, step_width - 1)
        except curses.error:
            pass
        idx_row += 1
    # return matrix
    return step_height, step_width, size_columns, size_rows


def compact_cpus(stdscr, pos_y, pos_x, width, jetson):
    _, _, _, size_rows = cpu_grid(stdscr, jetson.cpu['cpu'], cpu_gauge, pos_y, pos_x + 1, size_width=width - 2)
    return size_rows


class CPU(Page):

    def __init__(self, stdscr, jetson):
        super(CPU, self).__init__("CPU", stdscr, jetson)
        # List all chart CPU
        size_cpu = len(jetson.cpu['cpu'])
        self._chart_cpus = [Chart(jetson, str(idx + 1), self.update_chart, color_text=curses.COLOR_BLUE,
                                  color_chart=[curses.COLOR_BLUE]) for idx in range(size_cpu)]

    def update_chart(self, jetson, name):
        cpu = jetson.cpu['cpu'][int(name) - 1]
        return {
            'active': cpu['online'],
            'value': [100 - cpu.get("idle", 100)],
        }

    def print_cpu(self, stdscr, idx, cpu, pos_y, pos_x, size_h, size_w):
        # Print status CPU
        governor = cpu.get('governor', '').capitalize()
        label_chart_cpu = "{percent: >3.0f}% {governor}".format(percent=100 - cpu.get('idle', 100), governor=governor)
        # Print chart
        chart = self._chart_cpus[idx]
        chart.draw(stdscr, [pos_x, pos_x + size_w], [pos_y, pos_y + size_h - 2], label=label_chart_cpu, y_label=False)
        # Print model
        model = cpu['model'] if 'model' in cpu else ''
        model = model[:size_w]
        try:
            stdscr.addstr(pos_y + size_h - 1, pos_x, model, curses.A_NORMAL)
        except curses.error:
            pass
        # Print info
        if 'freq' in cpu:
            freq = cpu['freq']
            freq['online'] = cpu['online']
            freq['name'] = "Frq"
            try:
                freq_gauge(stdscr, pos_y + size_h, pos_x, size_w, cpu['freq'])
            except curses.error:
                pass

    def draw(self, key, mouse):
        # Screen size
        height, width, first = self.size_page()
        # Print gauge all CPU
        total = self.jetson.cpu['total']
        total['name'] = 'ALL'
        cpu_gauge(self.stdscr, 0, total, first + 1, 1, '', width)
        # Print all GRID CPU
        step_height, step_width, size_columns, size_rows = cpu_grid(
            self.stdscr, self.jetson.cpu['cpu'], self.print_cpu, first + 2, 1, size_height=height - 4, size_width=width - 8)
        # Print CPU Y axis
        chart = self._chart_cpus[0]
        for i in range(size_rows):
            chart.draw_y_axis(self.stdscr, first + 2 + i * step_height, 1 + step_width * size_columns, step_height - 3)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import curses
from .colors import NColors
from .common import size_to_string


header = [
    ("PID", {'clm': 7, 'fn': lambda x: str(x)}),
    ("USER", {'clm': 9, 'fn': lambda x: x}),
    ("GPU", {'clm': 5, 'fn': lambda x: x}),
    ("TYPE", {'clm': 6, 'fn': lambda x: x[0]}),
    ("PRI", {'clm': 5, 'fn': lambda x: str(x)}),
    ("S", {'clm': 4, 'fn': lambda x: x}),
    ("CPU%", {'clm': 7, 'fn': lambda x: "{:.1f}".format(x)}),
    ("MEM", {'clm': 8, 'fn': lambda x: size_to_string(x, 'k')}),
    ("GPU MEM", {'clm': 12, 'fn': lambda x: size_to_string(x, 'k')}),
    ("Command", {'clm': 20, 'fn': lambda x: x}),
]


class ProcessTable(object):

    def __init__(self, stdscr, jetson):
        self.stdscr = stdscr
        self.jetson = jetson
        self.line_sort = 8
        self.type_reverse = True

    def draw(self, pos_y, pos_x, width, height, key, mouse):
        # Plot low bar background line
        try:
            self.stdscr.addstr(pos_y, 0, " " * width, NColors.igreen())
        except curses.error:
            return 0
        title_counter = 0
        for idx, (title, info) in enumerate(header):
            try:
                # Check if pressed
                if mouse and mouse[1] == pos_y and title_counter <= mouse[0] <= title_counter + info['clm']:
                    if self.line_sort != idx:
                        self.line_sort = idx
                        self.type_reverse = True
                    else:
                        self.type_reverse = not self.type_reverse
                # Draw title
                title = "[{}]".format(title) if idx == self.line_sort else title
                self.stdscr.addstr(pos_y, title_counter, title, NColors.igreen() | curses.A_BOLD)
                title_counter += info['clm']
            except curses.error:
                break
        # Sort table for selected line
        try:
            sorted_processes = self.jetson.processes
            sorted_processes = sorted(sorted_processes, key=lambda x: x[self.line_sort], reverse=self.type_reverse)
        except IndexError:
            pass
        # Draw all processes
        # Instantiate the number of process variable to avoid an unbound local error if the process table is empty.
        nprocess = 0
        for nprocess, process in enumerate(sorted_processes):
            # Skip unit size process
            counter = 0
            for (value, (name, info)) in zip(process, header):
                # Print all values in a nice view
                try:
                    self.stdscr.addstr(pos_y + nprocess + 1, counter, info['fn'](value), curses.A_NORMAL)
                    counter += info['clm']
                except curses.error:
                    break
            # Stop loop if table is bigger than height
            if nprocess > height - 2:
                break
        return nprocess
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import re
import sys
import json
# from warnings import warn
from datetime import datetime, timedelta
from multiprocessing import Event, AuthenticationError
from threading import Thread
from .service import JtopManager
from .core.hardware import get_platform_variables
from .core.memory import Memory
from .core.fan import Fan
from .core.gpu import GPU
from .core.jetson_clocks import JetsonClocks
from .core.nvpmodel import NVPModel
from .core.common import compare_versions, get_var, get_local_interfaces, status_disk
from .core.jetson_libraries import get_libraries, get_cuda, get_opencv
from .core.exceptions import JtopException
# Fix connection refused for python 2.7
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError
try:
    PermissionError
except NameError:
    PermissionError = OSError
if sys.version_info[0] == 2:
    from socket import error as ConnectionRefusedError
# Create logger
logger = logging.getLogger(__name__)
# Gain timeout lost connection
TIMEOUT_GAIN = 3
# Version match
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)


class DateTimeEncoder(json.JSONEncoder):
    def default(self, z):
        if isinstance(z, datetime):
            return (str(z))
        elif isinstance(z, timedelta):
            return (str(z))
        else:
            return super().default(z)


class jtop(Thread):
    """
    This class control the access to your board, from here you can control your
    NVIDIA Jetson board or read the jetson_clocks status or change the nvp model.
    """

    def __init__(self, interval=1.0):
        """
        When you initialize your jtop you can setup a communication speed **interval**,
        if there is another jtop running this speed will be not used.

        When jtop is started you can read the server speed in **interval** property.

        :param interval: Interval to setup the jtop speed (in seconds), defaults to 1.0
        :type interval: float, optional
        """
        # Initialize Thread super class
        super(jtop, self).__init__()
        # Local Event thread
        self._trigger = Event()
        # Error message from thread
        self._error = None
        # Start server
        self._running = False
        # Load interval
        self._interval = float(interval)
        # Initialize observer
        self._observers = set()
        # Stats read from service
        self._stats = {}
        # Read stats
        JtopManager.register('get_queue')
        JtopManager.register("sync_data")
        JtopManager.register('sync_event')
        # Initialize broadcaster manager
        self._broadcaster = JtopManager()
        # Initialize board variable
        self._board = {}
        self._thread_libraries = Thread(target=self._load_jetson_libraries, args=[])
        self._thread_libraries.daemon = True
        self._thread_libraries.start()
        # Initialize gpu info
        self._gpu = GPU()
        # Initialize memory controller
        self._memory = Memory()
        # Initialize fan
        self._fan = Fan()
        # Load jetson_clocks status
        self._jetson_clocks = None
        # Load NV Power Mode
        self._nvpmodel = None

    def _load_jetson_libraries(self):
        # Load platform
        self._board['platform'] = get_platform_variables()
        # Load all variables
        cuda_version = get_cuda()
        opencv_version, opencv_cuda = get_opencv()
        os_variables = get_libraries()
        libraries = {
            'CUDA': cuda_version,
            'OpenCV': opencv_version,
            'OpenCV-Cuda': opencv_cuda,
        }
        libraries.update(os_variables)
        # Make dictionaries
        self._board['libraries'] = libraries
        # Loaded from script
        logger.debug("Loaded jetson_variables variables")

    def attach(self, observer):
        """
        Attach an observer to read the status of jtop. You can add more observer that you want.

        The function **must** be with this format:

        .. code-block:: python

            def observer(jetson):
                pass

        The input of your callback will be the jetson object.

        To detach a function, please look :func:`~detach`

        :param observer: The function to call
        :type observer: function
        """
        self._observers.add(observer)

    def detach(self, observer):
        """
        Detach an observer from jtop

        To attach a function, please look :func:`~attach`

        :param observer:  The function to detach
        :type observer: function
        """
        self._observers.discard(observer)

    def restore(self, max_counter=10):
        """
        This block method will restore all jtop configuration, in order:

        #. Switch off, disable on boot **jetson_clocks** and remove configuration file(reference :py:class:`~jtop.core.jetson_clocks.JetsonClocks`)
        #. Set all **fan** to default profile and all speed to zero (reference :py:class:`~jtop.core.fan.Fan`)
        #. Set to default **nvpmodel** (reference :py:class:`~jtop.core.nvpmodel.NVPModel`)
        #. **clear** the configuration jtop file

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    for status, message in jetson.restore():
                        if status:
                            print(message)
                        else:
                            print("Fail")

        :param max_counter: Counter time for each test before fail, defaults to 10
        :type max_counter: int, optional
        :yield: Generator of all operations to restore your NVIDIA Jetson
        :rtype: bool, str
        :raises JtopException: if the connection with the server is lost,
            not active or your user does not have the permission to connect to *jtop.service*
        """
        # Reset jetson_clocks
        if self.jetson_clocks is not None:
            # Disable jetson_clocks
            try:
                self.jetson_clocks = False
            except JtopException as ex:
                yield False, ex
            # Wait jetson_clocks boot
            counter = 0
            while self.ok() and (counter < max_counter):
                if not self.jetson_clocks:
                    break
                counter += 1
            yield counter != max_counter, "jetson_clocks disabled"
            # Disable jetson_clocks on boot
            self.jetson_clocks.boot = False
            # Wait jetson_clocks boot
            counter = 0
            while self.ok() and (counter < max_counter):
                if not self.jetson_clocks.boot:
                    break
                counter += 1
            yield counter != max_counter, "jetson_clocks disabled on boot"
            # Clear configuration
            self.jetson_clocks.clear_config()
            # Wait jetson_clocks clear configuration
            counter = 0
            while self.ok() and (counter < max_counter):
                if not self.jetson_clocks.is_config():
                    break
                counter += 1
            yield counter != max_counter, "clear jetson_clocks configuration file"
        # Reset fan control
        if self.fan is not None:
            # Reset mode fan
            for name in self.fan:
                profile_default = self.fan.get_profile_default(name)
                self.fan.set_profile(name, profile_default)
                counter = 0
                while self.ok() and (counter < max_counter):
                    if self.fan.get_profile(name) == profile_default:
                        break
                    counter += 1
                yield counter != max_counter, "Fan \"{name}\" set to profile \"{profile}\"".format(name=name, profile=profile_default)
            # Reset speed to zero
            for idx in range(len(self.fan[name]['speed'])):
                self.fan.set_speed(name, 0, idx)
                counter = 0
                while self.ok() and (counter < max_counter):
                    if self.fan.get_speed(name, idx) == 0:
                        break
                    counter += 1
                yield counter != max_counter, "Fan \"{name}[{idx}]\" set speed to 0".format(name=name, idx=idx)
        # Set to default nvpmodel
        if self.nvpmodel is not None:
            # Read default value
            default = self.nvpmodel.get_default()
            try:
                self.nvpmodel.set_nvpmodel_id(default['id'])
            except JtopException as ex:
                yield False, ex
            # Wait nvpmodel is default
            counter = 0
            while self.ok() and (counter < max_counter):
                if self.nvpmodel == default['name']:
                    break
                counter += 1
            yield counter != max_counter, "Default nvpmodel[{id}] {name}".format(id=default['id'], name=default['name'])
        # Clear config file
        self._controller.put({'config': 'reset'})
        yield True, "Clear jtop configuration file"

    @property
    def engine(self):
        """
        Engine status, in this property you can find like: APE, DLA, NVDEC, NVENC, and other

        The output of this property is a dictionary:

        * **name group X** - In this group are collected all engines similar
            * **name engine a** - In this key there is a dictionary with engine status
            * **name engine b** - Same like above, there is the dictionary status engine

        For each engine the dictionary is defined like the table below:

        ========== ================ ==============================================
        Name       Type             Description
        ========== ================ ==============================================
        online     :py:class:`bool` Status of the engine
        min        :py:class:`int`  Minimum frequency of the core in **kHz** :sup:`A`
        max        :py:class:`int`  Maximum frequency of the core in **kHz** :sup:`A`
        cur        :py:class:`int`  Current frequency of the core in **kHz**
        ========== ================ ==============================================

        .. note::

                Note **A**
                    Some engines doesn't have a *min* and *max* frequency

        :return: Dictionary of all active engines
        :rtype: dict
        """
        return self._stats.get('engines', {})

    @property
    def board(self):
        """
        Board status, in this property where are available all information about your device.

        The output is a dictionary with 3 keys:

        * **platform**
        * **hardware**
        * **libraries**

        *platform* (If you are running in docker this output is extracted from your **container**)

        ============= ====================================== ====================================================
        Name          Type                                   Description
        ============= ====================================== ====================================================
        Machine       :py:func:`platform.machine`            Mapped output: machine type
        System        :py:func:`platform.system`             Mapped output: Type system
        Distribution  :py:func:`platform.linux_distribution` Mapped output: Linux distribution
        Release       :py:func:`platform.release`            Mapped output: Release kernel
        Python        :py:func:`platform.python_version`     Mapped output: Python version is running jtop
        ============= ====================================== ====================================================

        *hardware* (If you are running in docker this output is extracted from your **host**)

        ===================== ====================================== ====================================================
        Name                  Type                                   Description
        ===================== ====================================== ====================================================
        Model                 :py:class:`str`                        Model name from :code:`/sys/firmware/devicetree/base/model`
        699-level Part Number :py:class:`str`                        699 part number (read reference)
        P-Number              :py:class:`str`                        Part number (read reference)
        BoardIDs              :py:class:`str`                        *(Optional)* Board ID in :code:`/proc/device-tree/nvidia,boardids`
        Module                :py:class:`str`                        Conversion from P-Number to Module name
        SoC                   :py:class:`str`                        System on Chip :code:`/proc/device-tree/compatible`
        CUDA Arch BIN         :py:class:`str`                        Cuda Architecture
        Codename              :py:class:`str`                        *(Optional)* Codename architecture
        Serial Number         :py:class:`str`                        Serial Number board :code:`/sys/firmware/devicetree/base/serial-number`
        L4T                   :py:class:`str`                        Linux 4 Tegra :sup:`A`
        Jetpack               :py:class:`str`                        From L4T is evaluated the Jetpack running on your NVIDIA Jetson
        ===================== ====================================== ====================================================

        *libraries* (If you are running in docker this output is extracted from your **container**)

        ===================== ====================================== ====================================================
        Name                  Type                                   Description
        ===================== ====================================== ====================================================
        CUDA                  :py:class:`str`                        CUDA version :sup:`B`
        OpenCV                :py:class:`str`                        OpenCV version use :code:`opencv_version`
        OpenCV-Cuda           :py:class:`bool`                       Check running :code:`opencv_version --verbose`
        cuDNN                 :py:class:`str`                        Check and read the version with :code:`dpkg -l`
        TensorRT              :py:class:`str`                        Check and read the version with :code:`dpkg -l`
        VPI                   :py:class:`str`                        Check and read the version with :code:`dpkg -l`
        Vulkan                :py:class:`str`                        Run and read :code:`which vulkaninfo`
        ===================== ====================================== ====================================================

        .. note::

            Note **A**
                The Linux For Tegra (L4T) is extracted in two ways:
                    1. Reading :code:`/etc/nv_tegra_release`
                    2. Reading version in **nvidia-l4t-core** package

            Note **B**
                The CUDA version is read depend of the version in:
                    1. On :code:`/usr/local/cuda/version.txt`
                    2. Running :code:`nvcc --version`

        .. admonition:: Reference

            #. `699 Part number - Before Jetpack 5 <https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-3243/index.html>`_
            #. `699 Part number - After Jetpack 5 <https://docs.nvidia.com/jetson/archives/r34.1/DeveloperGuide/text/HR/JetsonEepromLayout.html>`_

        :return: Status board, hardware and libraries
        :rtype: dict
        """
        # Wait thread end
        self._thread_libraries.join()
        # Return board status
        return self._board

    @property
    def fan(self):
        """
        Fan status and control. This property show speed, rpm and control every fan on your board

        The :py:class:`~jtop.core.fan.Fan` class is readable such a :py:class:`dict` where for each key are colleted all metrics for each Fan.

        ============= =================== ====================================================
        Name          Type                Description
        ============= =================== ====================================================
        speed         :py:class:`list`    List of speed between [0, 100]
        rpm           :py:class:`list`    *(Optional)* List of RPM for each fan
        profile       :py:class:`str`     Fan Profile, read :py:func:`~jtop.core.fan.Fan.all_profiles()`
        governor      :py:class:`str`     (Jetson with JP5+) Governor fan
        control       :py:class:`str`     (Jetson with JP5+) Type of controller
        ============= =================== ====================================================

        If you are working with Jetpack 5 or higher, the fan profile map nvfancontrol `nvfancontrol <https://docs.nvidia.com/jetson/archives/r34.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#fan-profile-control>`_

        you can also control your fan with:

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Print profile
                    print(jetson.fan.profile)
                    # Set new profile
                    jetson.fan.profile = 'quiet'
                    # Read speed
                    print(jetson.fan.speed)
                    # Set new speed
                    jetson.fan.speed = 90

        Full documentation on :py:class:`~jtop.core.fan.Fan`

        :return: Status Fan
        :rtype: Fan
        :raises JtopException: Wrong speed fan name, profile or speed
        """  # noqa
        return self._fan

    @property
    def nvpmodel(self):
        """
        The NVP Model control voltage regulators, and power tree to optimize power efficiency.
        It supports three optimized power budgets, such as 10 watts, 15 watts, and 30 watts.
        For each power budget, several configurations are possible with various CPU frequencies and number of cores online.

        Capping the memory, CPU, and GPU frequencies, and number of online CPU, GPU TPC, DLA and PVA cores at a prequalified level confines the module to the target mode.

        This method simplify in a set of functions and variables this controller.

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read current nvpmodel name
                    print(jetson.nvpmodel)
                    # List of all nvpmodel available
                    models = jetson.nvpmodel.models
                    print(models)
                    # You can write a string for a name
                    jetson.nvpmodel = models[0]
                    # or an the ID name is also allowed
                    jetson.nvpmodel = 0

        You can also increase/decrease the ID 

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    jetson.nvpmodel += 1
                    # or
                    jetson.nvpmodel = jetson.nvpmodel + 1

        Advanced features are available in :py:class:`~jtop.core.nvpmodel.NVPModel`

        .. admonition:: Reference

            #. `NVP Model - Jetson TX/Nano <https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-283/Tegra%20Linux%20Driver%20Package%20Development%20Guide/power_management_tx2.html#wwpID0E0AM0HA>`_
            #. `NVP Model - Jetson Xavier <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonXavierNxSeriesAndJetsonAgxXavierSeries.html#supported-modes-and-power-efficiency>`_
            #. `NVP Model - Jetson Orin <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#supported-modes-and-power-efficiency>`_

        :return: Return the name of NV Power Mode
        :rtype: NVPModel or None
        :raises JtopException: if the nvp model does not exist
        """  # noqa
        return self._nvpmodel

    @nvpmodel.setter
    def nvpmodel(self, value):
        if self._nvpmodel is None:
            raise JtopException("nvpmodel not available on this board")
        if isinstance(value, int):
            return self._nvpmodel.set_nvpmodel_id(value)
        elif isinstance(value, str):
            return self._nvpmodel.set_nvpmodel_name(value)
        else:
            ValueError("Data type not allowed {type}".format(type=type(value)))

    @property
    def jetson_clocks(self):
        """
        jetson_clocks is a tool provided for all NVIDIA Jetson to maximize all performance, read reference for more information.

        You can use this attribute like a simple boolean or use the advanced features in :py:class:`~jtop.core.jetson_clocks.JetsonClocks`

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Change status jetson_clocks
                    jetson.jetson_clocks = not jetson.jetson_clocks
                    # Set on board boot
                    jetson.jetson_clocks.boot = True
                    # Read status jetson_clocks service
                    print(jetson.jetson_clocks.status)

        .. admonition:: Reference

            #. `jetson_clocks - Jetson TX/Nano <https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-283/index.html#page/Tegra%2520Linux%2520Driver%2520Package%2520Development%2520Guide%2Fpower_management_tx2.html%23>`_
            #. `jetson_clocks - Jetson Xavier <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonXavierNxSeriesAndJetsonAgxXavierSeries.html>`_
            #. `jetson_clocks - Jetson Orin <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#maximizing-jetson-orin-performance>`_

        :return: status jetson_clocks script
        :rtype: JetsonClocks or None
        :raises ValueError: Wrong jetson_clocks value
        """  # noqa
        return self._jetson_clocks

    @jetson_clocks.setter
    def jetson_clocks(self, value):
        if self._jetson_clocks is None:
            raise JtopException("jetson_clocks not available on this board")
        self._jetson_clocks.set_enable(value)

    @property
    def stats(self):
        """
        This property return a simplified version of tegrastats,
        it is simple to use if you want log the NVIDIA Jetson status with pandas or in a csv file.

        This property is a simplified version of all data collected from your NVIDIA Jetson,
        if you need more detailed information, please use the other jtop properties

        The field listed are:

        ============= ============================== ========================= ====================================================
        Name          Type                           Reference                 Description
        ============= ============================== ========================= ====================================================
        time          :py:class:`datetime.datetime`                            local time in your board
        uptime        :py:class:`datetime.timedelta` :py:attr:`~uptime`        up time on your board
        cpu **X**     :py:class:`float`              :py:attr:`~cpu`           The status for each cpu in your board, if disabled *OFF*
        RAM           :py:class:`float`              :py:attr:`~memory`        RAM used / total
        SWAP          :py:class:`float`              :py:attr:`~memory`        SWAP used / total
        EMC           :py:class:`float`              :py:attr:`~memory`        *(Optional)* EMC Percentage of bandwidth
        IRAM          :py:class:`float`              :py:attr:`~memory`        *(Optional)* IRAM used / total
        GPU           :py:class:`float`              :py:attr:`~gpu`           *(Optional)* Status of your GPU
        engine **X**  :py:class:`float`              :py:attr:`~engine`        *(Optional)* Frequency for each engine, if disabled *OFF*
        fan           :py:class:`float`              :py:attr:`~fan`           *(Optional)* Fan speed
        Temp **X**    :py:class:`float`              :py:attr:`~power`         *(Optional)* Current power from rail X
        Temp **TOT**  :py:class:`float`              :py:attr:`~power`         *(Optional)* Total power
        jetson_clocks :py:class:`str`                :py:attr:`~jetson_clocks` *(Optional)* Status of jetson_clocks, human readable
        nvpmodel      :py:class:`str`                :py:attr:`~nvpmodel`      *(Optional)* NV Power Model name active
        ============= ============================== ========================= ====================================================

        :return: Compacts jetson statistics
        :rtype: dict
        """
        stats = {'time': datetime.now(), 'uptime': self.uptime}
        # -- CPU --
        for idx, cpu in enumerate(self.cpu['cpu']):
            stats["CPU{idx}".format(idx=idx + 1)] = 100 - int(cpu['idle']) if cpu['online'] else 'OFF'
        # -- MEMORY --
        tot_ram = self.memory['RAM']['tot']
        stats['RAM'] = self.memory['RAM']['used'] / tot_ram if tot_ram > 0 else 0
        tot_swap = self.memory['SWAP']['tot']
        stats['SWAP'] = self.memory['SWAP']['used'] / tot_swap if tot_swap > 0 else 0
        if 'EMC' in self.memory:
            stats['EMC'] = self.memory['EMC']['val']
        if 'IRAM' in self.memory:
            tot_iram = self.memory['IRAM']['tot']
            stats['IRAM'] = float(self.memory['IRAM']['used']) / tot_iram if tot_iram > 0 else 0
        # -- GPU --
        for idx, gpu in enumerate(self.gpu.values()):
            gpu_name = 'GPU' if idx == 0 else 'GPU{idx}'.format(idx=idx)
            stats[gpu_name] = gpu['status']['load']
        # -- Engines --
        for group in self.engine:
            for name, engine in self.engine[group].items():
                stats[name] = engine['cur'] if engine['online'] else 'OFF'
        # -- FAN --
        # Print all Fan
        for name, fan in self.fan.items():
            for idx, speed in enumerate(fan['speed']):
                stats['Fan {name}{idx}'.format(idx=idx, name=name)] = speed
        # -- Temperature --
        for temp in self.temperature:
            stats["Temp {name}".format(name=temp)] = self.temperature[temp]['temp']
        # -- Power --
        # Load all current power from each power rail
        if self.power:
            for name, rail in self.power['rail'].items():
                stats["Power {name}".format(name=name)] = rail['power']
            # Load total current power
            stats['Power TOT'] = self.power['tot']['power']
        # -- jetson_clocks --
        if self.jetson_clocks is not None:
            stats['jetson_clocks'] = 'ON' if self.jetson_clocks else 'OFF'
        # -- NV Power Model --
        if self.nvpmodel is not None:
            stats['nvp model'] = self.nvpmodel.name
        return stats

    def json(self, stats=False, **json_args):
        """
        This method export all metrics in a `json` readable output.

        You can export all metrics or the same output in :py:attr:`stats` depending of the parameter input.

        :param stats: json with same output of :py:attr:`stats`, defaults to False
        :type stats: bool, optional
        :param json_args: additional keyword arguments passed to json.dumps
        :type json_args: Any
        :return: json output requested
        :rtype: str
        """
        if stats:
            json_args.setdefault("cls", DateTimeEncoder)
            return json.dumps(self.stats, **json_args)
        # Read all variable and build a complete json
        return json.dumps(self._stats, **json_args)

    @property
    def memory(self):
        """
        This property show in a simple way all memories available, the main output is available in this way:

        * **RAM** - It is a dictionary with all information about RAM :sup:`A`
        * **SWAP** - It is a dictionary with all information about SWAP
        * **EMC** - It is a dictionary with EMC data, not in all boards this data is available
        * **IRAM** - It is a dictionary with SWAP data, not in all boards this data is available

        You can also use this property to set a new swap, deactivate or clear cache,
        read all methods available :py:class:`~jtop.core.memory.Memory`

        example:

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Print memory status
                    print(jetson.memory)
                    # make a new 10Gb swap
                    jetson.memory.swap_set(10, on_boot=False)


        For each dictionary there are specific outputs

        *RAM* :sup:`A`

        ========== =================== ====================================================
        Name       Type                Description
        ========== =================== ====================================================
        tot        :py:class:`int`     Total RAM in **KB**
        used       :py:class:`int`     Total used RAM in **KB**
        free       :py:class:`int`     Free RAM in **KB**
        buffers    :py:class:`int`     Buffered RAM in **KB**
        cached     :py:class:`int`     Cached RAM in **KB**
        shared     :py:class:`int`     Shared RAM in **KB**, for NVIDIA Jetson the RAM used from GPU
        lfb        :py:class:`int`     Large Free Block in **4MB**
        ========== =================== ====================================================

        *SWAP*

        ========== =================== ====================================================
        Name       Type                Description
        ========== =================== ====================================================
        tot        :py:class:`int`     Total SWAP in **KB**
        used       :py:class:`int`     Total used SWAP in **KB**
        cached     :py:class:`int`     Cached RAM in **KB**
        table      :py:class:`dict`    Dictionary with all swap available :sup:`B`
        ========== =================== ====================================================

        *EMC* (if available on your device)

        ========== =================== ====================================================
        Name       Type                Description
        ========== =================== ====================================================
        online     :py:class:`bool`    Status EMC
        val        :py:class:`int`     Percentage of bandwidth used relative to running frequency
        cur        :py:class:`int`     Current working frequency in **kHz**
        max        :py:class:`int`     Max EMC frequency usable in **kHz**
        min        :py:class:`int`     Min EMC frequency usable in **kHz**
        ========== =================== ====================================================

        *IRAM* (if available on your device)

        ========== =================== ====================================================
        Name       Type                Description
        ========== =================== ====================================================
        tot        :py:class:`int`     Total IRAM in **KB**
        used       :py:class:`int`     Total used IRAM in **KB**
        lfb        :py:class:`int`     Large Free Block in **4MB**
        ========== =================== ====================================================

        .. note::

            Note **A**
                The RAM is measured reading the file :code:`/proc/meminfo` for each field is read:

                ========== ====================================================
                Name       Description
                ========== ====================================================
                tot        :code:`MemTotal`
                used       :code:`MemTotal - (Buffers + Cached)`
                free       :code:`MemFree`
                buffers    :code:`Buffers`
                cached     :code:`Cached + SReclaimable`
                ========== ====================================================

            Note **B**
                The swap table is a list of dictionary with this data

                ========== =================== ==============================================
                Name       Type                Description
                ========== =================== ==============================================
                type       :py:class:`str`     Type of partition
                prio       :py:class:`int`     Priority partition
                size       :py:class:`int`     Size partition in **KB**
                used       :py:class:`int`     Used part of this partition in **KB**
                boot       :py:class:`bool`    Check if this swap start on boot
                ========== =================== ==============================================

        :return: memory status
        :rtype: Memory
        """
        return self._memory

    @property
    def cpu(self):
        """
        this property return a dictionary with all information for each core about frequency, idle, and other.

        This dictionary is made:

        * **total** - The aggregate values for all cores of (user, nice, system, idle)
        * **cpu** - a list with a dictionary for each core

        For each core the dictionary is defined:

        ========== ================= =======================================
        Name       Type              Description
        ========== ================= =======================================
        online     :py:class:`bool`  Status core
        governor   :py:class:`str`   Type of governor running on the core
        freq       :py:class:`dict`  Frequency of the core :sup:`A`
        info_freq  :py:class:`dict`  Frequency of the core :sup:`A`
        idle_state :py:class:`dict`  All Idle state running
        user       :py:class:`float` User percentage utilization :sup:`B`
        nice       :py:class:`float` Nice percentage utilization :sup:`B`
        system     :py:class:`float` System percentage utilization :sup:`B`
        idle       :py:class:`float` Idle percentage :sup:`B`
        model      :py:class:`str`   Model core running
        ========== ================= =======================================

        .. note::

            Note **A**
                The frequency dictionary is defined like below:

                ========== =================== ==============================================
                Name       Type                Description
                ========== =================== ==============================================
                min        :py:class:`int`     Minimum frequency of the core in **kHz**
                max        :py:class:`int`     Maximum frequency of the core in **kHz**
                cur        :py:class:`int`     Current frequency of the core in **kHz**
                ========== =================== ==============================================

            Note **B**
                If a core is offline, this data is not key is not available

        .. admonition:: Reference

            #. https://docs.kernel.org/admin-guide/pm/cpuidle.html
            #. https://www.linuxhowtos.org/System/procstat.htm
            #. https://rosettacode.org/wiki/Linux_CPU_utilization

        :return: A dictionary with the aggregate status and a list of all CPUs, described above.
        :rtype: dict
        """
        # Return CPU status
        return self._stats['cpu']

    @property
    def processes(self):
        """
        Return a list with all processing running in GPU.

        For each item are collected all information about the process.

        ========== ================= =======================================
        Item       Type              Description
        ========== ================= =======================================
        0          :py:class:`int`   PID process running
        1          :py:class:`str`   User start the process
        2          :py:class:`str`   GPU used :sup:`E`
        3          :py:class:`str`   Type of process :sup:`F`
        4          :py:class:`int`   Priority
        5          :py:class:`str`   State process :sup:`A`
        6          :py:class:`float` CPU percent utilization :sup:`B`
        7          :py:class:`int`   Memory occupied :sup:`C`
        8          :py:class:`int`   GPU Memory occupied :sup:`D`
        9          :py:class:`str`   Process name
        ========== ================= =======================================

        .. note::

            Note **A**
                This field indicating the status process:
                    * **R**  Running
                    * **S**  Sleeping in an interruptible wait
                    * **D**  Waiting in uninterruptible disk sleep
                    * **Z**  Zombie
                    * **T**  Stopped (on a signal)
                    * **t**  Tracing stop
                    * **X**  Dead

            Note **B**
                Measure the CPU percent utilization are read live the values from :code:`/proc/[PID]/stat`:
                    * **#14 utime** - CPU time spent in user code, measured in *clock ticks*
                    * **#15 stime** - CPU time spent in kernel code, measured in *clock ticks*
                    * **#22 starttime** - Time when the process started, measured in *clock ticks*

                Where *clock ticks* is `SC_CLK_TCK <http://pubs.opengroup.org/onlinepubs/009695399/utilities/getconf.html>`_

                It is also used :code:`/proc/uptime` to know the system up time.

                The CPU percent is:

                .. code-block:: python
                    :class: no-copybutton

                        total_time = utime + stime
                        total_time = uptime - (starttime / clock_ticks)
                        cpu_usage = 100 * (total_time / clock_ticks)

            Note **C**
                Extract resident set size (VmRSS) (Second field) in :code:`/proc/[PID]/statm`:
                    VmRSS is the resident set size of the process,
                    which is the portion of the process's memory that is held in RAM and is not swapped out to disk.
                    This is the amount of memory that the process is currently using.

            Note **D**
                This value is the GPU memory occupied from the process.

            Note **E**
                GPU used:
                    * **I**: Integrated GPU
                    * **dX**: Discrete GPU with number graphic card (next release)

            Note **F**
                Type of process:
                    * **Graphic**: Graphic process
                    * **System**: System process (next release)

        .. admonition:: Reference

            #. https://man7.org/linux/man-pages/man5/proc.5.html

        :return: list of all GPU processes
        :rtype: list
        """
        return self._stats['processes']

    @property
    def gpu(self):
        """
        This property show in a simple way all GPU available on your board.

        You can also use this attribute to enable/disable 3D scaling:

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # change 3D scaling status
                    jetson.gpu.set_scaling_3D = not jetson.gpu.set_scaling_3D

        The :py:class:`~jtop.core.gpu.GPU` class is readable such a :py:class:`dict` where for each key are colleted all metrics for each GPU.

        ============= =================== ====================================================
        Name          Type                Description
        ============= =================== ====================================================
        type          :py:class:`str`     Type of GPU (integrated, discrete)
        status        :py:class:`dict`    Status of GPU :sup:`A`
        freq          :py:class:`dict`    Frequency GPU :sup:`B`
        power_control :py:class:`dict`    *(Optional)* Type of power control
        ============= =================== ====================================================

        .. note::

            Note **A**
                Status GPU, from current load to features

                =========== =================== ==============================================
                Name        Type                Description
                =========== =================== ==============================================
                railgate    :py:class:`bool`    Status Railgate
                tpc_pg_mask :py:class:`bool`    Status TPC PG Mask (for NVP model)
                3d_scaling  :py:class:`bool`    Status 3D scaling
                load        :py:class:`float`   Current GPU load
                =========== =================== ==============================================

            Note **B**
                The frequency dictionary is defined like below:

                ========== =================== ==============================================
                Name       Type                Description
                ========== =================== ==============================================
                governor   :py:class:`str`     Name GPU governor
                min        :py:class:`int`     Minimum GPU frequency in **kHz**
                max        :py:class:`int`     Maximum GPU frequency in **kHz**
                cur        :py:class:`int`     Current GPU frequency in **kHz**
                GPC        :py:class:`list`    List GPC frequency in **kHz** (Available for Orin series)
                ========== =================== ==============================================

        :return: current status of your GPU.
        :rtype: GPU
        """
        # Extract GPU
        return self._gpu

    @property
    def power(self):
        """
        All NVIDIA Jetson have one ore more integrate three-channel `INA3221 <https://www.ti.com/product/INA3221>`_ to measure the power consumption.

        This jtop attribute collect and show the output from each rail in a simple view.

        ============= =================== ====================================================
        Name          Type                Description
        ============= =================== ====================================================
        rail          :py:class:`dict`    A dictionary with all thermal rails
        tot           :py:class:`dict`    Total estimate board power
        ============= =================== ====================================================

        The total power is the **sum of all rails** or

        * Jetson Xavier NX and Jetson Orin NX - `Output VDD_IN <https://forums.developer.nvidia.com/t/orin-nx-power-data-from-jtop/242804/5>`_
        * Jetson Nano - `Output POM_5V_IN <https://forums.developer.nvidia.com/t/power-consumption-monitoring/73608/8>`_

        For each rail there are different values available

        ============= =================== ====================================================
        Name          Type                Description
        ============= =================== ====================================================
        online        :py:class:`bool`    If sensor is online
        type          :py:class:`str`     Type of sensors (For NVIDIA Jetson is INA3221)
        status        :py:class:`str`     *(if available)* Status sensor
        volt          :py:class:`int`     Gets rail voltage in millivolts
        curr          :py:class:`int`     Gets rail current in milliamperes
        power         :py:class:`int`     Gets rail power in milliwatt
        avg           :py:class:`int`     Gets rail power average in milliwatt
        warn          :py:class:`int`     *(if available)* Gets rail average current limit in milliamperes
        crit          :py:class:`int`     *(if available)* Gets rail instantaneous current limit in milliamperes
        ============= =================== ====================================================

        .. admonition:: Reference

            #. `Power Consumption - Jetson TX/Nano <https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-283/Tegra%20Linux%20Driver%20Package%20Development%20Guide/power_management_tx2.html#wwpID0E0EE0HA>`_
            #. `Power Consumption - Jetson Xavier <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonXavierNxSeriesAndJetsonAgxXavierSeries.html#software-based-power-consumption-modeling>`_
            #. `Power Consumption - Jetson Orin <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#software-based-power-consumption-modeling>`_
            #. `INA3221 datasheet <https://www.ti.com/product/INA3221>`_

        :return: A dictionary with a list of power and the total
        :rtype: dict
        """  # noqa
        return self._stats['power']

    @property
    def temperature(self):
        """
        BSP thermal management features are part of the firmware running on BPMP for Jetson platforms running any host operating system (host OS) on the CPU.

        this attribute provide a dictionary with a list of all thermal rail available on your board.

        All temperatures are in Celsius

        ============= =================== ====================================================
        Name          Type                Description
        ============= =================== ====================================================
        online        :py:class:`bool`    If sensor is online
        temp          :py:class:`int`     Gets rail voltage in Celsius. *(If offline show -256)*
        max           :py:class:`int`     *(if available)* Gets rail average current limit in Celsius
        crit          :py:class:`int`     *(if available)* Gets rail instantaneous current limit in Celsius
        ============= =================== ====================================================

        .. note::

            all measures are with a 0.5 °C precision margin

        Not all values are available, jtop hide some values usually not available:

        * **PMIC** - For NVIDIA Jetson TX/Nano

        .. admonition:: Reference

            #. `Thermal specification - Jetson TX/Nano <https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-283/Tegra%20Linux%20Driver%20Package%20Development%20Guide/power_management_tx2.html#wwpID0E0IH0HA>`_
            #. `Thermal specification - Jetson Xavier <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonXavierNxSeriesAndJetsonAgxXavierSeries.html#thermal-management-in-bpmp>`_
            #. `Thermal specification - Jetson Orin <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#thermal-management-in-bpmp>`_

        :return: Temperature dictionary
        :rtype: dict
        """  # noqa
        return self._stats['temperature']

    @property
    def local_interfaces(self):
        """
        Local interfaces information and hostname

        This dictionary the status of your local network

        * **hostname** - Hostname board
        * **interfaces** - A dictionary with name and IP address for all interfaces listed

        :return: Local interfaces and hostname
        :rtype: dict
        """
        return get_local_interfaces()

    @property
    def disk(self):
        """
        Disk status properties, in dictionary are included

        * **total** - Total disk space in GB
        * **available** - Space available in GB
        * **used** - Disk space used in GB
        * **available_no_root**

        :return: Disk information
        :rtype: dict
        """
        return status_disk()

    @property
    def uptime(self):
        """
        Up time, The time since the board is turned on.

        This command read the value in :code:`/proc/uptime` .

        :return: Board up time
        :rtype: datetime.timedelta
        """
        return timedelta(seconds=self._stats['uptime'])

    def run(self):
        """ """
        # https://gist.github.com/schlamar/2311116
        # https://stackoverflow.com/questions/13074847/catching-exception-in-context-manager-enter
        try:
            while self._running:
                # Send alive message
                if self._controller.empty():
                    self._controller.put({})
                # Read stats from jtop service
                self._get_data()
        except Exception:
            # Store error message
            self._error = sys.exc_info()

    def _get_data(self):
        try:
            # Check if is not set event otherwise wait
            if not self._sync_event.is_set():
                self._sync_event.wait(self._interval * TIMEOUT_GAIN)
            # Read stats from jtop service
            data = self._sync_data.copy()
            if not data:
                raise JtopException("Error connection")
            # Clear event
            self._sync_event.clear()
        except EOFError:
            # Raise jtop exception
            raise JtopException("Lost connection with jtop server")
        # Decode and update all jtop data
        self._stats = data
        # -- GPU --
        self._gpu._update(self._stats['gpu'])
        # -- MEMORY --
        self._memory._update(self._stats['mem'])
        # -- FAN --
        self._fan._update(self._stats['fan'])
        # -- JETSON_CLOCKS --
        if 'jc' in self._stats:
            self._jetson_clocks._update(self._stats['jc'])
        # -- NVP Model --
        if 'nvp' in self._stats:
            self._nvpmodel._update(self._stats['nvp'])
        # Set trigger
        self._trigger.set()
        # Notify all observers
        for observer in self._observers:
            # Call all observer in list
            observer(self)

    def _get_configuration(self):
        while True:
            # Send configuration connection
            self._controller.put({'interval': self._interval})
            # Return configuration
            data = self._controller.get(self._interval * TIMEOUT_GAIN)
            if 'init' in data:
                return data['init']

    def start(self):
        """
        The start() function start your jtop and you can start to read the NVIDIA Jetson status.

        This method is **not** needed to close jtop if you have open jtop using `with` like:

        .. code-block:: python

            with jtop() as jetson:
                pass

        :raises JtopException: if the connection with the server is lost,
            not active or your user does not have the permission to connect to *jtop.service*
        """
        # Connected to broadcaster
        try:
            self._broadcaster.connect()
        except FileNotFoundError as e:
            if e.errno == 2 or e.errno == 111:  # Message error: 'No such file or directory' or 'Connection refused'
                raise JtopException("The jtop.service is not active. Please run:\nsudo systemctl restart jtop.service")
            elif e.errno == 13:  # Message error: 'Permission denied'
                raise JtopException("I can't access jtop.service.\nPlease logout or reboot this board.")
            else:
                raise FileNotFoundError(e)
        except ConnectionRefusedError as e:
            if e.errno == 111:  # Connection refused
                # When server is off but socket files exists in /run
                raise JtopException("The jtop.service is not active. Please run:\nsudo systemctl restart jtop.service")
            else:
                raise ConnectionRefusedError(e)
        except PermissionError as e:
            if e.errno == 13:  # Permission denied
                raise JtopException("I can't access jtop.service.\nPlease logout or reboot this board.")
            else:
                raise PermissionError(e)
        except ValueError:
            # https://stackoverflow.com/questions/54277946/queue-between-python2-and-python3
            raise JtopException("Mismatch of Python versions between library and service")
        except AuthenticationError:
            raise JtopException("Authentication with jetson-stats server failed")
        # Initialize synchronized data and condition
        self._controller = self._broadcaster.get_queue()
        self._sync_data = self._broadcaster.sync_data()
        self._sync_event = self._broadcaster.sync_event()
        # Initialize connection
        init = self._get_configuration()
        # Get jtop service version
        service_version = init.get('version', 'unknown')
        # Check version compatibility between client and server raise exception only if minor version is different
        if not compare_versions(service_version, get_var(VERSION_RE)):
            raise JtopException("""Mismatch version jtop service: [{service_version}] and client: [{client_version}]. Please run:\n
sudo systemctl restart jtop.service""".format(
                service_version=service_version,
                client_version=get_var(VERSION_RE)))
        # Load server speed
        self._server_interval = init['interval']
        # Load board information
        self._board['hardware'] = init['board']['hardware']
        # Initialize gpu controller
        self._gpu._initialize(self._controller)
        # Initialize memory controller
        self._memory._initialize(self._controller, init['memory'])
        # Initialize fan
        self._fan._initialize(self._controller, init['fan'])
        # Initialize jetson_clocks (if exist)
        if init['jc']:
            self._jetson_clocks = JetsonClocks(self._controller)
        # Init NVP model (if exist)
        if 'nvpmodel' in init:
            self._nvpmodel = NVPModel(self._controller, init['nvpmodel'])
        # Wait first value
        self._get_data()
        # Run thread reader
        self._running = True
        self.daemon = True
        super(jtop, self).start()

    @property
    def interval(self):
        """
        Speed jtop service. This speed can be different compare the speed specified in :func:`~jtop` constructor

        :return: jtop interval (in seconds)
        :rtype: float
        """
        return self._server_interval

    @property
    def interval_user(self):
        """
        This is the same speed specified in :func:`~jtop` constructor

        :return: jtop user interval (in seconds)
        :rtype: float
        """
        return self._interval

    def loop_for_ever(self):
        """
        This blocking method is needed when you design your python code to work only by callback.

        Before to run this method remember to attach a callback using :func:`~attach`

        A simple example to use this method is below

        .. code-block:: python

            def read_stats(jetson):
                stats = jetson.stats

            jetson = jtop()
            jetson.attach(read_stats)
            # Blocking method
            jetson.loop_for_ever()
        """
        self.start()
        # Blocking function to catch exceptions
        while self.ok():
            try:
                self.join(timeout=0.1)
            except (KeyboardInterrupt, SystemExit):
                # Close jtop
                self.close()

    def ok(self, spin=False):
        """
        This method is needed when you start jtop using `with` like below

        .. code-block:: python

            with jtop() as jetson:
                while jetson.ok():
                    stats = jetson.stats

        This method is usually blocking, and is not needed to add in your script a sleep function,
        when a new data will be available the function will release and you will read a new fresh data

        :param spin: If True, this function will be not blocking, defaults to False
        :type spin: bool, optional
        :raises ex_value: if jtop client fail, will be raise here
        :return: status jtop client
        :rtype: bool
        """
        # Wait if trigger is set
        if not spin:
            try:
                if not self._trigger.is_set():
                    if not self._trigger.wait(self._interval * TIMEOUT_GAIN):
                        self._running = False
            except (KeyboardInterrupt, SystemExit):
                self._running = False
        # Catch exception if exist
        if self._error:
            # Extract exception and raise
            ex_type, ex_value, tb_str = self._error
            ex_value.__traceback__ = tb_str
            raise ex_value
        # If there are not errors clear the event
        if self._running:
            self._trigger.clear()
        # Return the status
        return self._running

    def close(self):
        """
        This method close the jtop server.

        This method is **not** needed to close jtop if you have open jtop using `with` like:

        .. code-block:: python

            with jtop() as jetson:
                pass
        """
        # Wait thread end
        self._thread_libraries.join()
        # Switch off broadcaster thread
        self._running = False

    def __enter__(self):
        """ Enter function for 'with' statement """
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Exit function for 'with' statement """
        if exc_tb is not None:
            return False
        return True
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    @staticmethod
    def ok(message="OK"):
        return bcolors.OKGREEN + str(message) + bcolors.ENDC

    @staticmethod
    def warning(message="WARN"):
        return bcolors.WARNING + str(message) + bcolors.ENDC

    @staticmethod
    def fail(message="ERR"):
        return bcolors.FAIL + str(message) + bcolors.ENDC

    @staticmethod
    def bold(message):
        return bcolors.BOLD + str(message) + bcolors.ENDC
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
import time
import logging
# Launch command
from datetime import timedelta
from threading import Thread
# Local functions and classes
from .command import Command
from .common import get_uptime
from .exceptions import JtopException
# Create logger
logger = logging.getLogger(__name__)

PATH_JETSON_CLOCKS = ['/usr/bin/jetson_clocks', '/home/nvidia/jetson_clocks.sh']
COMMAND_TIMEOUT = 3.0
CONFIG_DEFAULT_BOOT = False
CONFIG_DEFAULT_DELAY = 60  # In seconds
CONFIG_DEFAULT_L4T_FILE = "l4t_dfs.conf"
# Type Engine
JC_ENGINES = re.compile(r'^(?P<name>[^ ]+) .* MaxFreq=(?P<frq>[^ ]+) .*')


def decode_show_message(lines):
    status = {}
    for line in lines:
        # Search configuration CPU config
        match = JC_ENGINES.search(line)
        if match:
            parsed_line = match.groupdict()
            # Extract name
            name = 'CPU' if 'cpu' in parsed_line['name'] else parsed_line['name']
            # Extrac max frequency
            freq = int(parsed_line['frq'])
            # add engine
            status[name] = freq
    return status


def jetson_clocks_alive(engines, data):
    for engine in engines:
        if engine == 'CPU':
            # Check minum and max frequency
            for cpu in data['cpu']['cpu']:
                if not cpu['online']:
                    continue
                cpu_frqs = cpu['freq']
                if cpu_frqs['max'] != cpu_frqs['min']:
                    return False
        elif engine == 'GPU':
            # Check minum and max frequency
            for _, gpu in data['gpu'].items():
                gpu_freqs = gpu['freq']
                if gpu_freqs['max'] != gpu_freqs['min']:
                    return False
        elif engine == 'EMC':
            # Check minum and max frequency
            # EMC check only if current frequency is the same of max
            emc = data['mem']['EMC']
            if 'max' not in emc:
                continue
            if emc['max'] != emc['cur']:
                return False
        else:
            # Find enging and check frequencies
            for _, engines_group in data['engines'].items():
                for name, engine_data in engines_group.items():
                    if engine == name and 'max' in engine_data:
                        # for all Engines check only if current frequency is the same of max
                        if engine_data['max'] != engine_data['cur']:
                            return False
    return True


class JetsonClocks(object):
    """
    jetson_clocks is a tool provided for all NVIDIA Jetson to maximize all performance, read reference for more information.

    With this class you can control when activate, deactivate or enable on boot.
    You can use in a simple way like a jtop attribute :py:attr:`~jtop.jtop.jetson_clocks`

    .. code-block:: python

        with jtop() as jetson:
            if jetson.ok():
                # Change status jetson_clocks
                jetson.jetson_clocks = not jetson.jetson_clocks
                # Set on board boot
                jetson.jetson_clocks.boot = True
                # Read status jetson_clocks service
                print(jetson.jetson_clocks.status)

    .. admonition:: Reference

        #. `jetson_clocks - Jetson TX/Nano <https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-283/index.html#page/Tegra%2520Linux%2520Driver%2520Package%2520Development%2520Guide%2Fpower_management_tx2.html%23>`_
        #. `jetson_clocks - Jetson Xavier <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonXavierNxSeriesAndJetsonAgxXavierSeries.html>`_
        #. `jetson_clocks - Jetson Orin <https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#maximizing-jetson-orin-performance>`_

    """ # noqa

    def __init__(self, controller):
        self._controller = controller
        self._enable = False
        self._thread = ''
        self._boot = False
        self._config = False

    def _update(self, jc_status):
        self._enable = jc_status['enable']
        self._thread = jc_status['thread']
        self._config = jc_status['config']
        self._boot = jc_status['boot']

    def set_enable(self, enable):
        """
        Enable jetson_clocks on your board. This method is equivalent to:

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Set a new state for jetson_clocks
                    jetson.jetson_clocks.set_enable(True)
                    # same of
                    jetson.jetson_clocks = True

        :param enable: Boolean status for enable and disable
        :type enable: bool
        :raises ValueError: if enable is not a boolean
        :raises JtopException: if jetson_clocks is in uncontrolled status
        """
        if not isinstance(enable, bool):
            raise ValueError("Use a boolean")
        if not self._config and not enable:
            raise JtopException("I cannot set jetson_clocks.\nPlease shutdown manually jetson_clocks")
        # Check if service is not started otherwise skip
        if self._thread in ['booting', 'activating', 'deactivating', 'uncontrolled']:
            return
        # Set new jetson_clocks configuration
        if enable is not self._enable:
            self._controller.put({'jc': {'enable': enable}})

    def get_enable(self):
        """
        Return current jetson_clocks status. This method is equivalent to:

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read jetson_clocks status
                    print(jetson.jetson_clocks.get_enable())
                    # same of
                    print(jetson.jetson_clocks)

        :return: Status jetson_clocks
        :rtype: bool
        """
        return self._enable

    def set_boot(self, value):
        """
        Enable jetson_clocks on board boot. This method is equivalent to:

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Set a new state for jetson_clocks
                    jetson.jetson_clocks.set_boot(True)
                    # same of
                    jetson.jetson_clocks.boot = True

        :param value: Boolean status for enable and disable jetson_clocks on boot
        :type value: bool
        :raises ValueError: if value is not a boolean
        """
        if not isinstance(value, bool):
            raise ValueError("Use a boolean")
        # Don't send a message if value is the same
        if value is self._boot:
            return
        # Set new jetson_clocks configuration
        self._controller.put({'jc': {'boot': value}})

    def get_boot(self):
        """
        Return if jetson_clocks start of board boot. This method is equivalent to:

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read jetson_clocks status
                    print(jetson.jetson_clocks.get_boot())
                    # same of
                    print(jetson.jetson_clocks.boot)

        :return: Status jetson_clocks on boot
        :rtype: bool
        """
        return self._boot

    @property
    def boot(self):
        """
        Same of :py:class:`~set_boot` and :py:class:`~get_boot` you can quickly enable and disable jetson_clocks.

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read jetson_clocks status
                    print(jetson.jetson_clocks.boot)
                    # Set a new state for jetson_clocks
                    jetson.jetson_clocks.boot = True

        :return: Status jetson_clocks on boot
        :rtype: bool
        """
        return self._boot

    @boot.setter
    def boot(self, value):
        self.set_boot(value)

    def is_config(self):
        """
        Return the jetson_clocks configuration status. If true jtop has stored a configuration file with all data

        :return: configuration stored
        :rtype: bool
        """
        return self._config

    def get_status(self):
        """
        Return the jetson_clocks service status:

        There are **six** options:

        ============== =============================================
        Status         Description
        ============== =============================================
        running        jetson_clocks is currently running
        booting        jetson_clocks is booting. This status appears in the first 60s when your board is started
        activating     jtop is running the service to activating jetson_clocks
        deactivating   jtop is running the service to deactivating jetson_clocks
        uncontrolled   jtop is not able to setup jetson_clocks. This message can appears if jetson_clocks was already running before jtop service started
        inactive       jetson_clocks is inactive
        ============== =============================================

        :return: status jetson_clocks service
        :rtype: str
        """
        # Check if is alive jetson_clocks
        if self._enable:
            return 'running'
        # Otherwise check status thread jetson_clocks
        return self._thread

    @property
    def status(self):
        """
        This property return the same output of :py:class:`~get_status`

        :return: status jetson_clocks service
        :rtype: str
        """
        return self.get_status()

    def clear_config(self):
        """
        This method clear the jetson_clocks configuration that use jtop.
        """
        # Clear jetson_clocks configuration
        self._controller.put({'jc': {'clear': True}})

    def __nonzero__(self):
        return self._enable

    def __bool__(self):
        return self._enable

    def __and__(self, other):
        return self._enable & bool(other)

    def __or__(self, other):
        return self._enable | bool(other)

    def __xor__(self, other):
        return self._enable ^ bool(other)

    def __int__(self):
        return int(self._enable)

    def __index__(self):
        return int(self._enable)

    def __eq__(self, other):
        if isinstance(other, bool):
            return self._enable == other
        elif isinstance(other, JetsonClocks):
            return self._enable == other._enable
        else:
            return False

    def __repr__(self):
        return str(self._enable)


class JetsonClocksService(object):
    """
        This controller manage the jetson_clocks service.
    """

    def __init__(self, config, fan):
        self._data = {}
        # Status thread enable/disable
        self._thread_status = 'inactive'
        # Initialization Jetson Clocks Object thread
        self._jetson_clocks_thread = None
        # List engines request
        self._engines_list = {}
        # Error message
        self._error = None
        # nvpmodel
        self._nvpmodel = None
        # Load configuration
        self._config = config
        # other modules
        self._fan = fan
        # Config file
        jetson_clocks_file = config.get('jetson_clocks', {}).get('l4t_file', CONFIG_DEFAULT_L4T_FILE)
        self._config_l4t = os.path.join(config.path, jetson_clocks_file)
        # Jetson Clocks path
        self._jetson_clocks_bin = ''
        for path in PATH_JETSON_CLOCKS:
            if os.path.exists(path):
                self._jetson_clocks_bin = path
                logger.info("jetson_clocks found in {cmd}".format(cmd=self._jetson_clocks_bin))
                break
        if not self._jetson_clocks_bin:
            logger.warning("jetson_clocks not available")

    def exists(self):
        return True if self._jetson_clocks_bin else False

    def initialization(self, nvpmodel, data):
        self._data = data
        self._nvpmodel = nvpmodel
        # List of all engines required
        self._engines_list = self.show()
        # Check if jetson_clocks is alive
        is_enabled = self.get_enable()
        # Check if exist configuration file
        if not os.path.isfile(self._config_l4t):
            if is_enabled:
                self._thread_status = 'uncontrolled'
                logger.warning("I can't store jetson_clocks configuration is already running!")
            else:
                self.store()
        # Check which version is L4T is loaded
        # if is before the 28.1 require to launch jetson_clock.sh only 60sec before the boot
        # https://devtalk.nvidia.com/default/topic/1027388/jetson-tx2/jetson_clock-sh-1-minute-delay/
        # Temporary disabled to find a best way to start this service.
        # The service on demand disabled doesn't improve the performance of the start-up
        # If jetson_clocks on boot run a thread
        config = self._config.get('jetson_clocks', {})
        if config.get('boot', CONFIG_DEFAULT_BOOT):
            logger.info("Starting jetson_clocks on boot")
            # Start thread Service client
            self.set_enable(True)

    def set_enable(self, enable, reset_nvpmodel=False):
        # If there are exception raise
        self._error_status()
        # Check status jetson_clocks
        running_status = self.is_running()
        if running_status != 'inactive':
            logger.warning("jetson_clocks is {status}".format(status=running_status))
            return False
        if enable:
            # Check if the configuration exist
            if not self.is_config():
                if not self.store():
                    return False
            # Start thread Service client
            self._jetson_clocks_thread = Thread(target=self._th_start, args=(reset_nvpmodel, ))
            self._jetson_clocks_thread.start()
        else:
            # Start thread Service client
            self._jetson_clocks_thread = Thread(target=self._th_stop, args=(reset_nvpmodel, ))
            self._jetson_clocks_thread.start()
        return True

    def get_enable(self):
        return jetson_clocks_alive(self._engines_list, self._data)

    def set_boot(self, value):
        # Extract configuration
        config = self._config.get('jetson_clocks', {})
        # Add new value
        config['boot'] = value
        # Set new jetson_clocks configuration
        self._config.set('jetson_clocks', config)

    def get_boot(self):
        config = self._config.get('jetson_clocks', {})
        return config.get('boot', CONFIG_DEFAULT_BOOT)

    def is_config(self):
        return os.path.isfile(self._config_l4t)

    def is_running(self):
        if self._jetson_clocks_thread is not None:
            if not self._jetson_clocks_thread.is_alive():
                self._thread_status = 'inactive'
        return self._thread_status

    def _fix_fan(self, status_fan):
        # Fix fan speed
        for name in status_fan:
            profile = status_fan[name]['profile']
            self._fan.set_profile(name, profile)
            logger.debug("Fan \"{name}\" profile: {profile}".format(name=name, profile=profile))
            if profile == 'manual':
                for idx, speed in enumerate(status_fan[name]['speed']):
                    self._fan.set_speed(name, speed, idx)
                    logger.debug("fan \"{name}\"[{idx}]: {speed}".format(name=name, idx=idx, speed=speed))

    def _th_start(self, reset_nvpmodel):
        self._thread_status = 'activating'
        # Check which version is L4T is loaded
        # if is before the 28.1 require to launch jetson_clock.sh only 60sec before the boot
        # https://devtalk.nvidia.com/default/topic/1027388/jetson-tx2/jetson_clock-sh-1-minute-delay/
        # Temporary disabled to find a best way to start this service.
        # The service on demand disabled doesn't improve the performance of the start-up
        # Measure remaining time from boot
        boot_time = timedelta(seconds=self._config.get('wait', CONFIG_DEFAULT_DELAY))
        up_time = timedelta(seconds=get_uptime())
        # If need time make a sleep
        if up_time < boot_time:
            self._thread_status = 'booting'
            delta = (boot_time - up_time).total_seconds()
            logger.info("Starting jetson_clocks in: {delta}s".format(delta=delta))
        # Status jetson_clocks
        logger.info("Start jetson_clocks with {status}".format(status=self._thread_status))
        if up_time < boot_time:
            time.sleep(delta)
        # Read fan speed
        status_fan = self._fan.get_status()
        if reset_nvpmodel and self._nvpmodel.exists():
            nvpmodel_id = self._nvpmodel.get_nvpmodel_id()
        # Start jetson_clocks
        Command.run_command([self._jetson_clocks_bin], repeat=5, timeout=COMMAND_TIMEOUT)
        # Fix fan speed
        self._fix_fan(status_fan)
        # Reset nvpmodel
        if reset_nvpmodel and self._nvpmodel.exists():
            self._nvpmodel.set_nvpmodel_id(nvpmodel_id, False)
        logger.info("jetson_clocks started")

    def _th_stop(self, reset_nvpmodel):
        self._thread_status = 'deactivating'
        logger.debug("Start jetson_clocks with {status}".format(status=self._thread_status))
        # Read fan speed
        status_fan = self._fan.get_status()
        if reset_nvpmodel and self._nvpmodel.exists():
            nvpmodel_id = self._nvpmodel.get_nvpmodel_id()
        # Run jetson_clocks
        Command.run_command([self._jetson_clocks_bin, '--restore', self._config_l4t], repeat=5, timeout=COMMAND_TIMEOUT)
        # Fix fan speed
        self._fix_fan(status_fan)
        # Reset nvpmodel
        if reset_nvpmodel and self._nvpmodel.exists():
            self._nvpmodel.set_nvpmodel_id(nvpmodel_id, False)
        logger.info("jetson_clocks stopped")

    def _error_status(self):
        # Catch exception if exist
        if self._error:
            # Extract exception and raise
            ex_type, ex_value, tb_str = self._error
            ex_value.__traceback__ = tb_str
            raise ex_value

    def close(self):
        # Switch off thread jetson clocks
        self._running = False
        # If jetson_clocks doesn't exist skip
        if not self.exists():
            return
        # If there are no thread running skip
        if self._jetson_clocks_thread is None:
            return
        if self._jetson_clocks_thread.is_alive():
            logger.warning("Wait switch off jetson_clocks thread")
            self._jetson_clocks_thread.join(COMMAND_TIMEOUT)

    def show(self):
        cmd = Command([self._jetson_clocks_bin, '--show'])
        list_engines = {}
        try:
            lines = cmd(timeout=COMMAND_TIMEOUT)
            list_engines = decode_show_message(lines)
        except Command.TimeoutException as e:
            logger.warning("Timeout {}".format(e))
        return list_engines

    def store(self):
        if self.is_config():
            logger.error("Configuration already stored in {file}".format(file=self._config_l4t))
            return False
        if self.get_enable():
            self._thread_status = 'uncontrolled'
            logger.error("jetson_clocks is running. I cannot store configuration")
            return False
        # Store configuration jetson_clocks
        cmd = Command([self._jetson_clocks_bin, '--store', self._config_l4t])
        try:
            message = cmd(timeout=COMMAND_TIMEOUT)
        except Command.CommandException:
            return False
        logger.info("Store jetson_clocks configuration in {file}".format(file=self._config_l4t))
        # Extract result
        return True if message else False

    def clear(self):
        if os.path.isfile(self._config_l4t):
            logger.info("Clear jetson_clocks config in {path}".format(path=self._config_l4t))
            # Remove configuration file
            os.remove(self._config_l4t)
            return True
        return False

    def get_status(self, data):
        self._data = data
        return {
            'enable': self.get_enable(),
            'thread': self.is_running(),
            'config': self.is_config(),
            'boot': self.get_boot(),
        }
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .jetson_variables import get_jetson_variables
from .common import cat
import os
import logging
import platform
# Load distro library from python3 or use platform
try:
    import distro
except ImportError:
    distro = platform
# Create logger
logger = logging.getLogger(__name__)


def get_parameter(path):
    if os.path.isfile(path):
        return cat(path).strip()


def get_platform_variables():
    return {
        'Machine': platform.machine(),
        'System': platform.system(),
        'Distribution': " ".join(distro.linux_distribution()),
        'Release': platform.release(),
        'Python': platform.python_version(),
    }


def get_x86_64_variables():
    hardware = {}
    hardware_path = "/sys/devices/virtual/dmi/id/"
    items = os.listdir(hardware_path)
    for item in sorted(items):
        if item in ['uevent', 'modalias', 'board_serial', 'bios_release', 'product_uuid', 'chassis_type']:
            continue
        path = os.path.join(hardware_path, item)
        output = ""
        if os.path.isfile(path):
            output = cat(path).strip()
        if not output or output == 'Default string':
            continue
        name = item.replace("_", " ").capitalize()
        hardware[name] = output
    return hardware


def get_hardware():
    # If hardware is ARM check if NVIDIA Jetson
    platform_board = platform.machine()
    logger.info("Hardware detected {}".format(platform_board))
    if platform_board == 'aarch64':
        # Load Jetson data
        jetson = get_jetson_variables()
        # Print main jetson variables
        if '699-level Part Number' in jetson:
            logger.info("NVIDIA Jetson 699-level Part Number={}".format(jetson['699-level Part Number']))
        else:
            logger.error("NVIDIA Jetson No 699-level Part Number detected!")
        if 'Module' in jetson:
            logger.info("NVIDIA Jetson Module={}".format(jetson['Module']))
        else:
            logger.error("NVIDIA Jetson No Module detected!")
        # Check L4T detection
        if jetson['L4T']:
            logger.info("NVIDIA Jetson detected L4T={}".format(jetson['L4T']))
        else:
            logger.error("NVIDIA Jetson L4T not detected!")
        return jetson
    elif platform_board == 'x86_64':
        return get_x86_64_variables()
    else:
        logger.warning("Unrecognized board {}".format(platform_board))
        return {}
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
import shlex
import curses
import sys
from pwd import getpwnam
import subprocess as sp
# Logging
import logging

from .gui import JTOPCONFIG
from .terminal_colors import bcolors
from .core.common import get_var
from .core.command import Command
from .core.jetson_variables import status_variables, install_variables
from .core.config import get_config_service
from .service import status_service, status_permission, install_service, set_service_permission
# Create logger
logger = logging.getLogger(__name__)
# Detect user
user = os.environ.get('USER', '')
# Get user from sudo
sudo_user = os.environ.get('SUDO_USER', '')
# If are both empty assign 'root'
user = sudo_user or 'root'
# Locate folder and type of installation
folder, _ = os.path.split(__file__)
folder = os.path.dirname(folder)
developer = os.path.isdir("{folder}/tests".format(folder=folder))
config = get_config_service()
# Version match
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
GUI_GRAPHIC_RE = re.compile(r'^AutomaticLoginEnable[ ]*=[ ]*[tT]rue')
REQUIRE_REBOOT = False

# -------------------- JTOP ------------------------------------------------


def update_jtop():
    cmd_update = Command(shlex.split('pip3 install --no-cache-dir -U jetson_stats'))
    cmd_update_msg = ""
    try:
        cmd_update_msg = cmd_update()
        cmd_update_msg = "JTOP updated!"
    except (OSError, Command.CommandException):
        cmd_update_msg = "Error to update jtop, please run:\n\nsudo pip3 install --no-cache-dir -U jetson_stats"
    return cmd_update_msg


def fix_service():
    copy = not developer
    # Install service (linking only for develop)
    install_service(folder, copy=copy)


def fix_variables():
    copy = not developer
    # Install variables
    install_variables(folder, copy=copy)


def fix_jtop_all():
    fix_service()
    set_service_permission()
    fix_variables()

# -------------------- Graphic ---------------------------------------------


def get_type_desktop():
    cmd_status_gui = Command(shlex.split('systemctl get-default'))
    status = 0
    try:
        status_gui = cmd_status_gui()[0]
        if status_gui == 'multi-user.target':
            status = 1
            if os.path.isfile('/etc/systemd/system/getty@tty1.service.d/autologin.conf'):
                status = 2
        elif status_gui == 'graphical.target':
            status = 3
            # Check GUI auto login
            if os.path.isfile('/etc/gdm3/custom.conf'):
                with open('/etc/gdm3/custom.conf', 'r') as f:
                    for line in f.readlines():
                        if GUI_GRAPHIC_RE.match(line):
                            status = 4
                            break
    except (OSError, Command.CommandException):
        pass
    return status


def desktop_is_type_one(data):
    return "[B1]" if data == 1 else " B1 "


def desktop_is_type_two(data):
    return "[B2]" if data == 2 else " B2 "


def desktop_is_type_tree(data):
    return "[B3]" if data == 3 else " B3 "


def desktop_is_type_four(data):
    return "[B4]" if data == 4 else " B4 "


def desktop_set_B1():
    global REQUIRE_REBOOT
    sp.call(shlex.split('systemctl set-default multi-user.target'))
    # sudo ln -fs /lib/systemd/system/getty@.service /etc/systemd/system/getty.target.wants/getty@tty1.service
    if os.path.isfile('/etc/systemd/system/getty.target.wants/getty@tty1.service'):
        os.remove('/etc/systemd/system/getty.target.wants/getty@tty1.service')
    os.symlink('/lib/systemd/system/getty@.service', '/etc/systemd/system/getty.target.wants/getty@tty1.service')
    # Remove autologin
    if os.path.isfile('/etc/systemd/system/getty@tty1.service.d/autologin.conf'):
        os.remove('/etc/systemd/system/getty@tty1.service.d/autologin.conf')
    REQUIRE_REBOOT = True


def desktop_set_B2():
    global REQUIRE_REBOOT
    sp.call(shlex.split('systemctl set-default multi-user.target'))
    # sudo ln -fs /lib/systemd/system/getty@.service /etc/systemd/system/getty.target.wants/getty@tty1.service
    if os.path.isfile('/etc/systemd/system/getty.target.wants/getty@tty1.service'):
        os.remove('/etc/systemd/system/getty.target.wants/getty@tty1.service')
    os.symlink('/lib/systemd/system/getty@.service', '/etc/systemd/system/getty.target.wants/getty@tty1.service')
    # If the file does not exist make it one
    if not os.path.isfile('/etc/systemd/system/getty@tty1.service.d/autologin.conf'):
        os.makedirs('/etc/systemd/system/getty@tty1.service.d')
    # Write file
    with open('/etc/systemd/system/getty@tty1.service.d/autologin.conf', 'w') as writer:
        writer.write("[Service]\nExecStart=\nExecStart=-/sbin/agetty --autologin {user} --noclear %I $TERM\n".format(user=user))
    # Change hownership
    pwnam = getpwnam(user)
    os.chown('/etc/systemd/system/getty@tty1.service.d/autologin.conf', pwnam.pw_uid, pwnam.pw_gid)
    REQUIRE_REBOOT = True


def desktop_set_B3():
    global REQUIRE_REBOOT
    sp.call(shlex.split('systemctl set-default graphical.target'))
    # sudo ln -fs /lib/systemd/system/getty@.service /etc/systemd/system/getty.target.wants/getty@tty1.service
    if os.path.isfile('/etc/systemd/system/getty.target.wants/getty@tty1.service'):
        os.remove('/etc/systemd/system/getty.target.wants/getty@tty1.service')
    os.symlink('/lib/systemd/system/getty@.service', '/etc/systemd/system/getty.target.wants/getty@tty1.service')
    # Remove autologin
    if os.path.isfile('/etc/systemd/system/getty@tty1.service.d/autologin.conf'):
        os.remove('/etc/systemd/system/getty@tty1.service.d/autologin.conf')
    # Update desktop login
    sp.call(shlex.split("update_desktop_login \"False\" {user}".format(user=user)))
    REQUIRE_REBOOT = True


def desktop_set_B4():
    global REQUIRE_REBOOT
    sp.call(shlex.split('systemctl set-default graphical.target'))
    # sudo ln -fs /lib/systemd/system/getty@.service /etc/systemd/system/getty.target.wants/getty@tty1.service
    if os.path.isfile('/etc/systemd/system/getty.target.wants/getty@tty1.service'):
        os.remove('/etc/systemd/system/getty.target.wants/getty@tty1.service')
    os.symlink('/lib/systemd/system/getty@.service', '/etc/systemd/system/getty.target.wants/getty@tty1.service')
    # Remove autologin
    if os.path.isfile('/etc/systemd/system/getty@tty1.service.d/autologin.conf'):
        os.remove('/etc/systemd/system/getty@tty1.service.d/autologin.conf')
    # Update desktop login
    sp.call(shlex.split("update_desktop_login \"True\" {user}".format(user=user)))
    REQUIRE_REBOOT = True


# ---------------------- Pages ---------------------------------------------
JTOP_MENU = {
    'title': 'status jtop',
    'description': 'config folder: {config}\njtop path:{jtop_path}'.format(config=config, jtop_path=folder),
    'menu': [
        (None, fix_jtop_all, "Fix all"),
        (status_service, fix_service, "Fix jetson-stats service"),
        (status_permission, set_service_permission, "Fix permissions for '{user}'".format(user=user)),
        (status_variables, fix_variables, "Fix jtop variables"),
    ],
}
DISPLAY_MENU = {
    'title': 'GUI menu option',
    'run_before': get_type_desktop,
    'menu': [
        (desktop_is_type_one, desktop_set_B1, "Text console, requiring user to login"),
        (desktop_is_type_two, desktop_set_B2, "Text console, automatically logged in as '{user}' user".format(user=user)),
        (desktop_is_type_tree, desktop_set_B3, "Desktop GUI, requiring user to login"),
        (desktop_is_type_four, desktop_set_B4, "Desktop GUI, automatically logged in as '{user}' user".format(user=user)),
    ]
}
INFORMATION_MENU = {
    'title': 'jtop {version} - main page'.format(version=get_var(VERSION_RE)),
    'description': "This tool provide a set of tool to configure your NVIDIA Jetson\
\n\nScripts: jtop, jetson_config, jetson_release, jetson_swap\nServices: jtop.service",
}
MAIN_PAGE = {
    'title': 'jetson-stats {version} - main page'.format(version=get_var(VERSION_RE)),
    'menu': [
        ("Health", JTOP_MENU, "Check the status of jetson-stats"),
        ("Update", update_jtop, "Update jetson-stats to the latest version"),
        ("Desktop", DISPLAY_MENU, "Enable/Disable boot from desktop"),
        ("About", INFORMATION_MENU, "Information about this configuration tool"),
    ]}


def jtop_config():
    # Check if running a root
    if os.getuid() == 0:
        # Run wrapper
        curses.wrapper(JTOPCONFIG, JTOP_MENU)
        sys.exit(0)
    # Quit with error
    print(bcolors.fail("Please run with sudo"))
    sys.exit(1)


def main():
    # Check if running a root
    if os.getuid() == 0:
        # Run wrapper
        curses.wrapper(JTOPCONFIG, MAIN_PAGE)
        if REQUIRE_REBOOT:
            print(bcolors.fail("require reboot"))
        sys.exit(0)
    # Quit with error
    print(bcolors.fail("Please run with sudo"))
    sys.exit(1)


if __name__ == "__main__":
    main()
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .common import cat, check_file
import os
# Logging
import logging
# Create logger
logger = logging.getLogger(__name__)


def total_power(power):
    """
    Function to measure the total watt

    :return: Total power and a second dictionary with all other measures
    :rtype: dict, dict
    """
    # In according with:
    # https://forums.developer.nvidia.com/t/power-consumption-monitoring/73608/8
    # https://github.com/rbonghi/jetson_stats/issues/51
    # https://forums.developer.nvidia.com/t/tegrastats-monitoring/217088/4?u=user62045
    # https://forums.developer.nvidia.com/t/orin-nx-power-data-from-jtop/242804/5
    total_name = ""
    for val in power:
        if val in ["POM_5V_IN", "VDD_IN"]:
            total_name = val
            break
    # Extract the total from list
    # Otherwise sum all values
    # Example for Jetson Xavier
    # https://forums.developer.nvidia.com/t/xavier-jetson-total-power-consumption/81016
    if total_name:
        total = power[total_name]
        total['name'] = total_name
        del power[total_name]
        return total, power
    # Otherwise measure all total power
    total = {'power': 0, 'avg': 0}
    for value in power.values():
        total['power'] += value['power']
        total['avg'] += value['avg']
    return total, power


def find_driver_power_folders(path):
    subdirectories = []
    driver_items = os.listdir(path)
    for item in driver_items:
        # All folder starting with:
        # - hwmon are JP5 compatible
        # - iio:device from JP4 or below
        if os.path.isdir(os.path.join(path, item)):
            driver_path = "{base_path}/{item}".format(base_path=path, item=item)
            if 'hwmon' in item:
                # Bring the second hwmon folder
                hwmon_name = os.listdir(driver_path)[0]
                driver_path = "{base_path}/{item}".format(base_path=driver_path, item=hwmon_name)
                subdirectories += [driver_path]
            elif 'iio:device' in item:
                subdirectories += [driver_path]
    return subdirectories


def find_all_i2c_power_monitor(i2c_path):
    power_sensor = {}
    if not os.path.isdir(i2c_path):
        logger.error("Folder {root_dir} doesn't exist".format(root_dir=i2c_path))
        return power_sensor
    power_i2c_sensors = {}
    items = os.listdir(i2c_path)
    for item in items:
        # Decode full path
        path = "{base_path}/{item}".format(base_path=i2c_path, item=item)
        name_path = "{path}/name".format(path=path)
        if os.path.isfile(name_path):
            raw_name = cat(name_path).strip()
            # Find all shunt and bus voltage monitor mounted on board
            # https://www.ti.com/product/INA3221
            if 'ina3221' in raw_name:
                # Check which type of driver is working
                power_i2c_sensors[item] = find_driver_power_folders(path)
    # Build list of all power folder outputs
    # Find all voltage and current monitor
    for name, paths in power_i2c_sensors.items():
        for path in paths:
            sensors = list_all_i2c_ports(path)
            power_sensor.update(sensors)
    if power_sensor:
        logger.info("Found I2C power monitor")
    return power_sensor


def read_power_status(data):
    values = {}
    power_type = data['type']
    try:
        for name, path in data.items():
            if 'type' in name:
                continue
            elif 'status' in name:
                values[name] = cat(path).strip()
                continue
            elif 'online' in name:
                values[name] = cat(path).strip() == '1'
                continue
            # Fix from values with "ma" in the end, like
            # warn 32760 ma
            raw_value = int(cat(path).split(" ")[0])
            values[name] = raw_value // 1000 if power_type != 'INA3221' else raw_value
    except OSError:
        values = {}
    return values


def list_all_i2c_ports(path):
    sensor_name = {}
    # Build list label and path
    for item in os.listdir(path):
        power_label_path = "{path}/{item}".format(path=path, item=item)
        # Check if there is a label
        if item.endswith("_label"):
            # Decode name for JP 5 or above
            raw_name = cat(power_label_path).strip()
            # Remove NC power (Orin family)
            # https://docs.nvidia.com/jetson/archives/r34.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#jetson-agx-orin-series
            if 'NC' in raw_name:
                logger.warn("Skipped NC {path}".format(path=power_label_path))
                continue
            # Build list current and average power read
            number_port = int(item.split("_")[0].strip("in"))
            # Skip "sum of shunt voltages" always number 7
            if number_port == 7:
                logger.warn("Skipped \"sum of shunt voltages\" {path}".format(path=power_label_path))
                continue
            # Build list of path
            warnings = {'type': 'INA3221'}
            if check_file("{path}/curr{num}_crit_alarm".format(path=path, num=number_port)):
                warnings['crit_alarm'] = "{path}/curr{num}_crit_alarm".format(path=path, num=number_port)
            if check_file("{path}/curr{num}_max_alarm".format(path=path, num=number_port)):
                warnings['max_alarm'] = "{path}/curr{num}_max_alarm".format(path=path, num=number_port)
            values = read_power_status(warnings)
            logger.info("Alarms {name} - {data}".format(name=raw_name, data=values))
            # Read Voltage, current and limits
            sensor = {'type': 'INA3221'}
            if check_file("{path}/in{num}_input".format(path=path, num=number_port)):
                sensor['volt'] = "{path}/in{num}_input".format(path=path, num=number_port)  # Voltage in mV
            if check_file("{path}/curr{num}_input".format(path=path, num=number_port)):
                sensor['curr'] = "{path}/curr{num}_input".format(path=path, num=number_port)  # Current in mA
            if check_file("{path}/curr{num}_max".format(path=path, num=number_port)):
                sensor['warn'] = "{path}/curr{num}_max".format(path=path, num=number_port)  # in mA
            if check_file("{path}/curr{num}_crit".format(path=path, num=number_port)):
                sensor['crit'] = "{path}/curr{num}_crit".format(path=path, num=number_port)  # in mA
            # If there is an file is added in list
            if len(sensor) > 1:
                sensor_name[raw_name] = sensor
        elif item.startswith("rail_name_"):
            # Decode name for JP 4 or previous
            raw_name = cat(power_label_path).strip()
            # Build list current and average power read
            number_port = int(item.lstrip("rail_name_"))
            # Build list of path
            # https://forums.developer.nvidia.com/t/jetson-tx1-ina226-power-monitor-with-i2c-interface/43819/5
            sensor = {'type': 'INA3221'}
            if check_file("{path}/in_voltage{num}_input".format(path=path, num=number_port)):
                sensor['volt'] = "{path}/in_voltage{num}_input".format(path=path, num=number_port)  # Voltage in mV
            if check_file("{path}/in_current{num}_input".format(path=path, num=number_port)):
                sensor['curr'] = "{path}/in_current{num}_input".format(path=path, num=number_port)  # Current in mA
            if check_file("{path}/in_power{num}_input".format(path=path, num=number_port)):
                sensor['power'] = "{path}/in_power{num}_input".format(path=path, num=number_port)  # Current in mW
            if check_file("{path}/warn_current_limit_{num}".format(path=path, num=number_port)):
                sensor['warn'] = "{path}/warn_current_limit_{num}".format(path=path, num=number_port)  # in mA
            if check_file("{path}/crit_current_limit_{num}".format(path=path, num=number_port)):
                sensor['crit'] = "{path}/crit_current_limit_{num}".format(path=path, num=number_port)  # in mA
            # If there is an file is added in list
            if len(sensor) > 1:
                sensor_name[raw_name] = sensor
    return sensor_name


def find_all_system_monitor(system_monitor):
    sensor_name = {}
    if not os.path.isdir(system_monitor):
        logger.error("Folder {root_dir} doesn't exist".format(root_dir=system_monitor))
        return sensor_name
    # Find all system power monitor
    for folder in os.listdir(system_monitor):
        local_path = "{path}/{folder}".format(path=system_monitor, folder=folder)
        name = folder.replace("ucsi-source-psy-", "")
        # Read type
        path_type = "{path}/type".format(path=local_path)
        type_supply = cat(path_type).strip() if check_file(path_type) else "SYSTEM"
        # Read model name
        path_name = "{path}/model_name".format(path=local_path)
        model_name = cat(path_name).strip() if check_file(path_name) else "<EMPTY>"
        # Find power, current and voltage
        sensor = {'type': type_supply}
        # Status power
        if check_file("{path}/online".format(path=local_path)):
            sensor['online'] = "{path}/online".format(path=local_path)
        if check_file("{path}/status".format(path=local_path)):
            sensor['status'] = "{path}/status".format(path=local_path)
        if check_file("{path}/voltage_now".format(path=local_path)):
            sensor['volt'] = "{path}/voltage_now".format(path=local_path)  # Voltage in mV
        if check_file("{path}/current_now".format(path=local_path)):
            sensor['curr'] = "{path}/current_now".format(path=local_path)  # Current in mA
        if check_file("{path}/current_max".format(path=local_path)):
            sensor['warn'] = "{path}/current_max".format(path=local_path)  # in mA
        # If there is an file is added in list
        if 'volt' in sensor and 'curr' in sensor:
            sensor_name[name] = sensor
            logger.info("Found name={name} type={type} model={model} in {path}".format(
                name=name, type=type_supply, model=model_name, path=folder))
        else:
            logger.warning("Skipped {name} type={type} in={path}".format(name=name, type=type_supply, path=folder))
    if sensor_name:
        logger.info("Found SYSTEM power monitor")
    return sensor_name


class PowerService(object):

    def __init__(self):
        self._power_sensor = {}
        self._power_avg = {}
        # Find all I2C sensors on board
        i2c_path = "/sys/bus/i2c/devices"
        system_monitor = "/sys/class/power_supply"
        if os.getenv('JTOP_TESTING', False):
            i2c_path = "/fake_sys/bus/i2c/devices"
            system_monitor = "/fake_sys/class/power_supply"
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=i2c_path))
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=system_monitor))
        # Load all power sensors
        self._power_sensor = find_all_i2c_power_monitor(i2c_path)
        self._power_sensor.update(find_all_system_monitor(system_monitor))
        if not self._power_sensor:
            logger.warning("Power sensors not found!")
        # Sort all power sensors
        self._power_sensor = dict(sorted(self._power_sensor.items(), key=lambda item: item[0]))

    def reset_avg_power(self):
        # Reset dictionary
        self._power_avg = {}

    def get_status(self):
        # If there are no sensors return an empty list
        if not self._power_sensor:
            return {}
        # Otherwise measure all values
        rails = {}
        for name, sensors in self._power_sensor.items():
            # Read status sensors
            values = read_power_status(sensors)
            if not values:
                continue
            # Measure power
            if 'power' not in values:
                power = values['volt'] * (float(values['curr']) / 1000)
                values['power'] = int(power)
            # if online not in group (for I2C)
            if 'online' not in values:
                values['online'] = True
            # Measure average Power between first and previous interval
            if name in self._power_avg:
                old_avg = self._power_avg[name]
                values['avg'] = (values['power'] + old_avg) // 2
                self._power_avg[name] = values['avg']
            else:
                values['avg'] = values['power']
                self._power_avg[name] = values['power']
            # Add on power status
            rails[name] = values
        # Measure total power
        total, rails = total_power(rails)
        return {'rail': rails, 'tot': total}
# EOF
